# BSD 2-Clause License
#
# Apprise - Push Notification Library.
# Copyright (c) 2025, Chris Caron <lead2gold@gmail.com>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

import base64
import contextlib
import mimetypes
import os
import time

from .. import exception
from ..common import ContentLocation
from ..locale import gettext_lazy as _
from ..url import URLBase
from ..utils.parse import parse_bool


class AttachBase(URLBase):
    """This is the base class for all supported attachment types."""

    # For attachment type detection; this amount of data is read into memory
    # 128KB (131072B)
    max_detect_buffer_size = 131072

    # Unknown mimetype
    unknown_mimetype = "application/octet-stream"

    # Our filename when we can't otherwise determine one
    unknown_filename = "apprise-attachment"

    # Our filename extension when we can't otherwise determine one
    unknown_filename_extension = ".obj"

    # The strict argument is a flag specifying whether the list of known MIME
    # types is limited to only the official types registered with IANA. When
    # strict is True, only the IANA types are supported; when strict is False
    # (the default), some additional non-standard but commonly used MIME types
    # are also recognized.
    strict = False

    # The maximum file-size we will accept for an attachment size. If this is
    # set to zero (0), then no check is performed
    # 1 MB = 1048576 bytes
    # 5 MB = 5242880 bytes
    # 1 GB = 1048576000 bytes
    max_file_size = 1048576000

    # By default all attachments types are inaccessible.
    # Developers of items identified in the attachment plugin directory
    # are requried to set a location
    location = ContentLocation.INACCESSIBLE

    # Here is where we define all of the arguments we accept on the url
    # such as: schema://whatever/?overflow=upstream&format=text
    # These act the same way as tokens except they are optional and/or
    # have default values set if mandatory. This rule must be followed
    template_args = {
        "cache": {
            "name": _("Cache Age"),
            "type": "int",
            # We default to (600) which means we cache for 10 minutes
            "default": 600,
        },
        "mime": {
            "name": _("Forced Mime Type"),
            "type": "string",
        },
        "name": {
            "name": _("Forced File Name"),
            "type": "string",
        },
        "verify": {
            "name": _("Verify SSL"),
            # SSL Certificate Authority Verification
            "type": "bool",
            # Provide a default
            "default": True,
        },
    }

    def __init__(self, name=None, mimetype=None, cache=None, **kwargs):
        """Initialize some general logging and common server arguments that
        will keep things consistent when working with the configurations that
        inherit this class.

        Optionally provide a filename to over-ride name associated with the
        actual file retrieved (from where-ever).

        The mime-type is automatically detected, but you can over-ride this by
        explicitly stating what it should be.

        By default we cache our responses so that subsiquent calls does not
        cause the content to be retrieved again.  For local file references
        this makes no difference at all.  But for remote content, this does
        mean more then one call can be made to retrieve the (same) data.  This
        method can be somewhat inefficient if disabled.  Only disable caching
        if you understand the consequences.

        You can alternatively set the cache value to an int identifying the
        number of seconds the previously retrieved can exist for before it
        should be considered expired.
        """

        super().__init__(**kwargs)

        if not mimetypes.inited:
            # Ensure mimetypes has been initialized
            mimetypes.init()

        # Attach Filename (does not have to be the same as path)
        self._name = name

        # The mime type of the attached content.  This is detected if not
        # otherwise specified.
        self._mimetype = mimetype

        # The detected_mimetype, this is only used as a fallback if the
        # mimetype wasn't forced by the user
        self.detected_mimetype = None

        # The detected filename by calling child class. A detected filename
        # is always used if no force naming was specified.
        self.detected_name = None

        # Absolute path to attachment
        self.download_path = None

        # Track open file pointers
        self.__pointers = set()

        # Set our cache flag; it can be True, False, None, or a (positive)
        # integer... nothing else
        if cache is not None:
            try:
                self.cache = cache if isinstance(cache, bool) else int(cache)

            except (TypeError, ValueError):
                err = f"An invalid cache value ({cache}) was specified."
                self.logger.warning(err)
                raise TypeError(err) from None

            # Some simple error checking
            if self.cache < 0:
                err = f"A negative cache value ({cache}) was specified."
                self.logger.warning(err)
                raise TypeError(err)

        else:
            self.cache = None

        # Validate mimetype if specified
        if self._mimetype and (
                next(
                    (
                        t
                        for t in mimetypes.types_map.values()
                        if self._mimetype == t
                    ),
                    None,
                )
                is None):
            err = f"An invalid mime-type ({mimetype}) was specified."
            self.logger.warning(err)
            raise TypeError(err)

        return

    @property
    def path(self):
        """Returns the absolute path to the filename.

        If this is not known or is know but has been considered expired (due to
        cache setting), then content is re-retrieved prior to returning.
        """

        if not self.exists():
            # we could not obtain our path
            return None

        return self.download_path

    @property
    def name(self):
        """Returns the filename."""
        if self._name:
            # return our fixed content
            return self._name

        if not self.exists():
            # we could not obtain our name
            return None

        if not self.detected_name:
            # If we get here, our download was successful but we don't have a
            # filename based on our content.
            ext = mimetypes.guess_extension(self.mimetype)
            self.detected_name = (
                f"{self.unknown_filename}"
                f"{ext if ext else self.unknown_filename_extension}"
            )

        return self.detected_name

    @property
    def mimetype(self):
        """Returns mime type (if one is present).

        Content is cached once determied to prevent overhead of future calls.
        """
        if not self.exists():
            # we could not obtain our attachment
            return None

        if self._mimetype:
            # return our pre-calculated cached content
            return self._mimetype

        if not self.detected_mimetype:
            # guess_type() returns: (type, encoding) and sets type to None
            # if it can't otherwise determine it.
            with contextlib.suppress(TypeError):
                # Directly reference _name and detected_name to prevent
                # recursion loop (as self.name calls this function)
                self.detected_mimetype, _ = mimetypes.guess_type(
                    self._name if self._name else self.detected_name,
                    strict=self.strict,
                )

        # Return our mime type
        return (
            self.detected_mimetype
            if self.detected_mimetype
            else self.unknown_mimetype
        )

    def exists(self, retrieve_if_missing=True):
        """Simply returns true if the object has downloaded and stored the
        attachment AND the attachment has not expired."""
        if self.location == ContentLocation.INACCESSIBLE:
            # our content is inaccessible
            return False

        cache = (
            self.template_args["cache"]["default"]
            if self.cache is None
            else self.cache
        )

        try:
            if (
                self.download_path
                and os.path.isfile(self.download_path)
                and cache
            ):

                # We have enough reason to look further into our cached content
                # and verify it has not expired.
                if cache is True:
                    # return our fixed content as is; we will always cache it
                    return True

                # Verify our cache time to determine whether we will get our
                # content again.
                age_in_sec = time.time() - os.stat(self.download_path).st_mtime
                if age_in_sec <= cache:
                    return True

        except OSError:
            # The file is not present
            pass

        return False if not retrieve_if_missing else self.download()

    def base64(self, encoding="ascii"):
        """Returns the attachment object as a base64 string otherwise None is
        returned if an error occurs.

        If encoding is set to None, then it is not encoded when returned
        """
        if not self:
            # We could not access the attachment
            self.logger.error(
                f"Could not access attachment {self.url(privacy=True)}."
            )
            raise exception.AppriseFileNotFound("Attachment Missing")

        try:
            with self.open() as f:
                # Prepare our Attachment in Base64
                return (
                    base64.b64encode(f.read()).decode(encoding)
                    if encoding
                    else base64.b64encode(f.read())
                )

        except (FileNotFoundError):
            # We no longer have a path to open
            raise exception.AppriseFileNotFound("Attachment Missing") from None

        except (TypeError, OSError) as e:
            self.logger.warning(
                "An I/O error occurred while reading {}.".format(
                    self.name if self else "attachment"
                )
            )
            self.logger.debug(f"I/O Exception: {e!s}")
            raise exception.AppriseDiskIOError(
                "Attachment Access Error") from e

    def invalidate(self):
        """Release any temporary data that may be open by child classes.
        Externally fetched content should be automatically cleaned up when this
        function is called.

        This function should also reset the following entries to None:
          - detected_name : Should identify a human readable filename
          - download_path: Must contain a absolute path to content
          - detected_mimetype: Should identify mimetype of content
        """

        # Remove all open pointers
        while self.__pointers:
            self.__pointers.pop().close()

        self.detected_name = None
        self.download_path = None
        self.detected_mimetype = None
        return

    def download(self):
        """This function must be over-ridden by inheriting classes.

        Inherited classes MUST populate:
          - detected_name: Should identify a human readable filename
          - download_path: Must contain a absolute path to content
          - detected_mimetype: Should identify mimetype of content

        If a download fails, you should ensure these values are set to None.
        """
        raise NotImplementedError(
            "download() is implimented by the child class."
        )

    def open(self, mode="rb"):
        """Return our file pointer and track it (we'll auto close later)"""
        pointer = open(self.path, mode=mode)  # noqa: SIM115
        self.__pointers.add(pointer)
        return pointer

    def chunk(self, size=5242880):
        """A Generator that yield chunks of a file with the specified size.

        By default the chunk size is set to 5MB (5242880 bytes)
        """

        with self.open() as file:
            while True:
                chunk = file.read(size)
                if not chunk:
                    break

                yield chunk

    def __enter__(self):
        """Support with keyword."""
        return self.open()

    def __exit__(self, value_type, value, traceback):
        """Stub to do nothing; but support exit of with statement
        gracefully."""
        return

    @staticmethod
    def parse_url(url, verify_host=True, mimetype_db=None, sanitize=True):
        """Parses the URL and returns it broken apart into a dictionary.

        This is very specific and customized for Apprise.

        Args:
            url (str): The URL you want to fully parse.
            verify_host (:obj:`bool`, optional): a flag kept with the parsed
                 URL which some child classes will later use to verify SSL
                 keys (if SSL transactions take place).  Unless under very
                 specific circumstances, it is strongly recomended that
                 you leave this default value set to True.

        Returns:
            A dictionary is returned containing the URL fully parsed if
            successful, otherwise None is returned.
        """

        results = URLBase.parse_url(
            url, verify_host=verify_host, sanitize=sanitize
        )

        if not results:
            # We're done; we failed to parse our url
            return results

        # Allow overriding the default config mime type
        if "mime" in results["qsd"]:
            results["mimetype"] = (
                results["qsd"].get("mime", "").strip().lower()
            )

        # Allow overriding the default file name
        if "name" in results["qsd"]:
            results["name"] = results["qsd"].get("name", "").strip().lower()

        # Our cache value
        if "cache" in results["qsd"]:
            # First try to get it's integer value
            try:
                results["cache"] = int(results["qsd"]["cache"])

            except (ValueError, TypeError):
                # No problem, it just isn't an integer; now treat it as a bool
                # instead:
                results["cache"] = parse_bool(results["qsd"]["cache"])

        return results

    def __len__(self):
        """Returns the filesize of the attachment."""
        if not self:
            return 0

        try:
            return os.path.getsize(self.path) if self.path else 0

        except OSError:
            # OSError can occur if the file is inaccessible
            return 0

    def __bool__(self):
        """Allows the Apprise object to be wrapped in an based 'if statement'.

        True is returned if our content was downloaded correctly.
        """
        return bool(self.path)

    def __del__(self):
        """Perform any house cleaning."""
        self.invalidate()



# BSD 2-Clause License
#
# Apprise - Push Notification Library.
# Copyright (c) 2025, Chris Caron <lead2gold@gmail.com>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

# For this plugin to work correct, the FCM server must be set up to allow
# for remote connections.

# Firebase Cloud Messaging
# Visit your console page: https://console.firebase.google.com
# 1. Create a project if you haven't already.  If you did the
#    {project} ID will be listed as name-XXXXX.
# 2. Click on your project from here to open it up.
# 3. Access your Web API Key by clicking on:
#     - The (gear-next-to-project-name) > Project Settings > Cloud Messaging

# Visit the following site to get you're Project information:
#    - https://console.cloud.google.com/project/_/settings/general/
#
# Docs: https://firebase.google.com/docs/cloud-messaging/send-message

# Legacy Docs:
# https://firebase.google.com/docs/cloud-messaging/http-server-ref\
#       #send-downstream
#
# If you Generate a new private key, it will provide a .json file
# You will need this in order to send an apprise messag
from json import dumps

import requests

from ...apprise_attachment import AppriseAttachment
from ...common import NotifyImageSize, NotifyType
from ...locale import gettext_lazy as _
from ...utils.logic import dict_full_update
from ...utils.parse import parse_bool, parse_list, validate_regex
from ..base import NotifyBase
from .color import FCMColorManager
from .common import FCM_MODES, FCMMode
from .priority import FCM_PRIORITIES, FCMPriorityManager

# Default our global support flag
NOTIFY_FCM_SUPPORT_ENABLED = False

try:
    from .oauth import GoogleOAuth

    # We're good to go
    NOTIFY_FCM_SUPPORT_ENABLED = True

except ImportError:
    # cryptography is the dependency of the .oauth library

    # Create a dummy object for init() call to work
    class GoogleOAuth:
        pass


# Our lookup map
FCM_HTTP_ERROR_MAP = {
    400: "A bad request was made to the server.",
    401: "The provided API Key was not valid.",
    404: "The token could not be registered.",
}


class NotifyFCM(NotifyBase):
    """A wrapper for Google's Firebase Cloud Messaging Notifications."""

    # Set our global enabled flag
    enabled = NOTIFY_FCM_SUPPORT_ENABLED

    requirements = {
        # Define our required packaging in order to work
        "packages_required": "cryptography"
    }

    # The default descriptive name associated with the Notification
    service_name = "Firebase Cloud Messaging"

    # The services URL
    service_url = "https://firebase.google.com"

    # The default protocol
    secure_protocol = "fcm"

    # A URL that takes you to the setup/help of the specific protocol
    setup_url = "https://github.com/caronc/apprise/wiki/Notify_fcm"

    # Project Notification
    # https://firebase.google.com/docs/cloud-messaging/send-message
    notify_oauth2_url = (
        "https://fcm.googleapis.com/v1/projects/{project}/messages:send"
    )

    notify_legacy_url = "https://fcm.googleapis.com/fcm/send"

    # There is no reason we should exceed 5KB when reading in a JSON file.
    # If it is more than this, then it is not accepted.
    max_fcm_keyfile_size = 5000

    # Allows the user to specify the NotifyImageSize object
    image_size = NotifyImageSize.XY_256

    # The maximum length of the body
    body_maxlen = 1024

    # Define object templates
    templates = (
        # OAuth2
        "{schema}://{project}/{targets}?keyfile={keyfile}",
        # Legacy Mode
        "{schema}://{apikey}/{targets}",
    )

    # Define our template
    template_tokens = dict(
        NotifyBase.template_tokens,
        **{
            "apikey": {
                "name": _("API Key"),
                "type": "string",
                "private": True,
            },
            "keyfile": {
                "name": _("OAuth2 KeyFile"),
                "type": "string",
                "private": True,
            },
            "project": {
                "name": _("Project ID"),
                "type": "string",
            },
            "target_device": {
                "name": _("Target Device"),
                "type": "string",
                "map_to": "targets",
            },
            "target_topic": {
                "name": _("Target Topic"),
                "type": "string",
                "prefix": "#",
                "map_to": "targets",
            },
            "targets": {
                "name": _("Targets"),
                "type": "list:string",
                "required": True,
            },
        },
    )

    template_args = dict(
        NotifyBase.template_args,
        **{
            "to": {
                "alias_of": "targets",
            },
            "mode": {
                "name": _("Mode"),
                "type": "choice:string",
                "values": FCM_MODES,
                "default": FCMMode.Legacy,
            },
            "priority": {
                "name": _("Mode"),
                "type": "choice:string",
                "values": FCM_PRIORITIES,
            },
            "image_url": {
                "name": _("Custom Image URL"),
                "type": "string",
            },
            "image": {
                "name": _("Include Image"),
                "type": "bool",
                "default": False,
                "map_to": "include_image",
            },
            # Color can either be yes, no, or a #rrggbb (
            # rrggbb without hashtag is accepted to)
            "color": {
                "name": _("Notification Color"),
                "type": "string",
                "default": "yes",
            },
        },
    )

    # Define our data entry
    template_kwargs = {
        "data_kwargs": {
            "name": _("Data Entries"),
            "prefix": "+",
        },
    }

    def __init__(
        self,
        project,
        apikey,
        targets=None,
        mode=None,
        keyfile=None,
        data_kwargs=None,
        image_url=None,
        include_image=False,
        color=None,
        priority=None,
        **kwargs,
    ):
        """Initialize Firebase Cloud Messaging."""
        super().__init__(**kwargs)

        if mode is None:
            # Detect our mode
            self.mode = FCMMode.OAuth2 if keyfile else FCMMode.Legacy

        else:
            # Setup our mode
            self.mode = (
                NotifyFCM.template_tokens["mode"]["default"]
                if not isinstance(mode, str)
                else mode.lower()
            )
            if self.mode and self.mode not in FCM_MODES:
                msg = f"The FCM mode specified ({mode}) is invalid."
                self.logger.warning(msg)
                raise TypeError(msg)

        # Used for Legacy Mode; this is the Web API Key retrieved from the
        # User Panel
        self.apikey = None

        # Path to our Keyfile
        self.keyfile = None

        # Our Project ID is required to verify against the keyfile
        # specified
        self.project = None

        # Initialize our Google OAuth module we can work with
        self.oauth = GoogleOAuth(
            user_agent=self.app_id,
            timeout=self.request_timeout,
            verify_certificate=self.verify_certificate,
        )

        if self.mode == FCMMode.OAuth2:
            # The project ID associated with the account
            self.project = validate_regex(project)
            if not self.project:
                msg = f"An invalid FCM Project ID ({project}) was specified."
                self.logger.warning(msg)
                raise TypeError(msg)

            if not keyfile:
                msg = "No FCM JSON KeyFile was specified."
                self.logger.warning(msg)
                raise TypeError(msg)

            # Our keyfile object is just an AppriseAttachment object
            self.keyfile = AppriseAttachment(asset=self.asset)
            # Add our definition to our template
            self.keyfile.add(keyfile)
            # Enforce maximum file size
            self.keyfile[0].max_file_size = self.max_fcm_keyfile_size

        else:  # Legacy Mode

            # The apikey associated with the account
            self.apikey = validate_regex(apikey)
            if not self.apikey:
                msg = f"An invalid FCM API key ({apikey}) was specified."
                self.logger.warning(msg)
                raise TypeError(msg)

        # Acquire Device IDs to notify
        self.targets = parse_list(targets)

        # Our data Keyword/Arguments to include in our outbound payload
        self.data_kwargs = {}
        if isinstance(data_kwargs, dict):
            self.data_kwargs.update(data_kwargs)

        # Include the image as part of the payload
        self.include_image = include_image

        # A Custom Image URL
        # FCM allows you to provide a remote https?:// URL to an image_url
        # located on the internet that it will download and include in the
        # payload.
        #
        # self.image_url() is reserved as an internal function name; so we
        # jsut store it into a different variable for now
        self.image_src = image_url

        # Initialize our priority
        self.priority = FCMPriorityManager(self.mode, priority)

        # Initialize our color
        self.color = FCMColorManager(color, asset=self.asset)
        return

    @property
    def access_token(self):
        """Generates a access_token based on the keyfile provided."""
        keyfile = self.keyfile[0]
        if not keyfile:
            # We could not access the keyfile
            self.logger.error(
                f"Could not access FCM keyfile {keyfile.url(privacy=True)}."
            )
            return None

        if not self.oauth.load(keyfile.path):
            self.logger.error(
                f"FCM keyfile {keyfile.url(privacy=True)} could not be loaded."
            )
            return None

        # Verify our project id against the one provided in our keyfile
        if self.project != self.oauth.project_id:
            self.logger.error(
                f"FCM keyfile {keyfile.url(privacy=True)} identifies itself"
                " for a different project"
            )
            return None

        # Return our generated key; the below returns None if a token could
        # not be acquired
        return self.oauth.access_token

    def send(self, body, title="", notify_type=NotifyType.INFO, **kwargs):
        """Perform FCM Notification."""

        if not self.targets:
            # There is no one to email; we're done
            self.logger.warning("There are no FCM devices or topics to notify")
            return False

        if self.mode == FCMMode.OAuth2:
            access_token = self.access_token
            if not access_token:
                # Error message is generated in access_tokengen() so no reason
                # to additionally write anything here
                return False

            headers = {
                "User-Agent": self.app_id,
                "Content-Type": "application/json",
                "Authorization": f"Bearer {access_token}",
            }

            # Prepare our notify URL
            notify_url = self.notify_oauth2_url

        else:  # FCMMode.Legacy
            headers = {
                "User-Agent": self.app_id,
                "Content-Type": "application/json",
                "Authorization": f"key={self.apikey}",
            }

            # Prepare our notify URL
            notify_url = self.notify_legacy_url

        # Acquire image url
        image = (
            self.image_url(notify_type)
            if not self.image_src
            else self.image_src
        )

        has_error = False
        # Create a copy of the targets list
        targets = list(self.targets)
        while len(targets):
            recipient = targets.pop(0)

            if self.mode == FCMMode.OAuth2:
                payload = {
                    "message": {
                        "token": None,
                        "notification": {
                            "title": title,
                            "body": body,
                        },
                    }
                }

                if self.color:
                    # Acquire our color
                    payload["message"]["android"] = {
                        "notification": {"color": self.color.get(notify_type)}
                    }

                if self.include_image and image:
                    payload["message"]["notification"]["image"] = image

                if self.data_kwargs:
                    payload["message"]["data"] = self.data_kwargs

                if recipient[0] == "#":
                    payload["message"]["topic"] = recipient[1:]
                    self.logger.debug(
                        "FCM recipient %s parsed as a topic", recipient[1:]
                    )

                else:
                    payload["message"]["token"] = recipient
                    self.logger.debug(
                        "FCM recipient %s parsed as a device token", recipient
                    )

            else:  # FCMMode.Legacy
                payload = {
                    "notification": {
                        "notification": {
                            "title": title,
                            "body": body,
                        }
                    }
                }

                if self.color:
                    # Acquire our color
                    payload["notification"]["notification"]["color"] = (
                        self.color.get(notify_type)
                    )

                if self.include_image and image:
                    payload["notification"]["notification"]["image"] = image

                if self.data_kwargs:
                    payload["data"] = self.data_kwargs

                if recipient[0] == "#":
                    payload["to"] = f"/topics/{recipient}"
                    self.logger.debug(
                        "FCM recipient %s parsed as a topic", recipient[1:]
                    )

                else:
                    payload["to"] = recipient
                    self.logger.debug(
                        "FCM recipient %s parsed as a device token", recipient
                    )

            # A more advanced dict.update() that recursively includes
            # sub-dictionaries as well
            dict_full_update(payload, self.priority.payload())

            self.logger.debug(
                "FCM %s POST URL: %s (cert_verify=%r)",
                self.mode,
                notify_url,
                self.verify_certificate,
            )
            self.logger.debug("FCM %s Payload: %s", self.mode, str(payload))

            # Always call throttle before any remote server i/o is made
            self.throttle()
            try:
                r = requests.post(
                    notify_url.format(project=self.project),
                    data=dumps(payload),
                    headers=headers,
                    verify=self.verify_certificate,
                    timeout=self.request_timeout,
                )
                if r.status_code not in (
                    requests.codes.ok,
                    requests.codes.no_content,
                ):
                    # We had a problem
                    status_str = NotifyBase.http_response_code_lookup(
                        r.status_code, FCM_HTTP_ERROR_MAP
                    )

                    self.logger.warning(
                        "Failed to send {} FCM notification: "
                        "{}{}error={}.".format(
                            self.mode,
                            status_str,
                            ", " if status_str else "",
                            r.status_code,
                        )
                    )

                    self.logger.debug("Response Details:\r\n%s", r.content)

                    has_error = True

                else:
                    self.logger.info("Sent %s FCM notification.", self.mode)

            except requests.RequestException as e:
                self.logger.warning(
                    "A Connection error occurred sending FCM notification."
                )
                self.logger.debug("Socket Exception: %s", str(e))

                has_error = True

        return not has_error

    @property
    def url_identifier(self):
        """Returns all of the identifiers that make this URL unique from
        another simliar one.

        Targets or end points should never be identified here.
        """
        return (self.secure_protocol, self.mode, self.apikey, self.project)

    def url(self, privacy=False, *args, **kwargs):
        """Returns the URL built dynamically based on specified arguments."""

        # Define any URL parameters
        params = {
            "mode": self.mode,
            "image": "yes" if self.include_image else "no",
            "color": str(self.color),
        }

        if self.priority:
            # Store our priority if one was defined
            params["priority"] = str(self.priority)

        if self.keyfile:
            # Include our keyfile if specified
            params["keyfile"] = NotifyFCM.quote(
                self.keyfile[0].url(privacy=privacy), safe=""
            )

        if self.image_src:
            # Include our image path as part of our URL payload
            params["image_url"] = self.image_src

        # Extend our parameters
        params.update(self.url_parameters(privacy=privacy, *args, **kwargs))

        # Add our data keyword/args into our URL response
        params.update({f"+{k}": v for k, v in self.data_kwargs.items()})

        reference = (
            NotifyFCM.quote(self.project)
            if self.mode == FCMMode.OAuth2
            else self.pprint(self.apikey, privacy, safe="")
        )

        return "{schema}://{reference}/{targets}?{params}".format(
            schema=self.secure_protocol,
            reference=reference,
            targets="/".join([NotifyFCM.quote(x) for x in self.targets]),
            params=NotifyFCM.urlencode(params),
        )

    def __len__(self):
        """Returns the number of targets associated with this notification."""
        return len(self.targets)

    @staticmethod
    def parse_url(url):
        """Parses the URL and returns enough arguments that can allow us to re-
        instantiate this object."""
        results = NotifyBase.parse_url(url, verify_host=False)
        if not results:
            # We're done early as we couldn't load the results
            return results

        # The apikey/project is stored in the hostname
        results["apikey"] = NotifyFCM.unquote(results["host"])
        results["project"] = results["apikey"]

        # Get our Device IDs
        results["targets"] = NotifyFCM.split_path(results["fullpath"])

        # Get our mode
        results["mode"] = results["qsd"].get("mode")

        # The 'to' makes it easier to use yaml configuration
        if "to" in results["qsd"] and len(results["qsd"]["to"]):
            results["targets"] += NotifyFCM.parse_list(results["qsd"]["to"])

        # Our Project ID
        if "project" in results["qsd"] and results["qsd"]["project"]:
            results["project"] = NotifyFCM.unquote(results["qsd"]["project"])

        # Our Web API Key
        if "apikey" in results["qsd"] and results["qsd"]["apikey"]:
            results["apikey"] = NotifyFCM.unquote(results["qsd"]["apikey"])

        # Our Keyfile (JSON)
        if "keyfile" in results["qsd"] and results["qsd"]["keyfile"]:
            results["keyfile"] = NotifyFCM.unquote(results["qsd"]["keyfile"])

        # Our Priority
        if "priority" in results["qsd"] and results["qsd"]["priority"]:
            results["priority"] = NotifyFCM.unquote(results["qsd"]["priority"])

        # Our Color
        if "color" in results["qsd"] and results["qsd"]["color"]:
            results["color"] = NotifyFCM.unquote(results["qsd"]["color"])

        # Boolean to include an image or not
        results["include_image"] = parse_bool(
            results["qsd"].get(
                "image", NotifyFCM.template_args["image"]["default"]
            )
        )

        # Extract image_url if it was specified
        if "image_url" in results["qsd"]:
            results["image_url"] = NotifyFCM.unquote(
                results["qsd"]["image_url"]
            )
            if "image" not in results["qsd"]:
                # Toggle default behaviour if a custom image was provided
                # but ONLY if the `image` boolean was not set
                results["include_image"] = True

        # Store our data keyword/args if specified
        results["data_kwargs"] = results["qsd+"]

        return results



# BSD 2-Clause License
#
# Apprise - Push Notification Library.
# Copyright (c) 2025, Chris Caron <lead2gold@gmail.com>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

# For this to work correctly you need to have a valid SFR DMC service account
# to whicthe API password can be generated. A "space" is also necessary
# (space = a logical separation between clients), which will give you a
# specific spaceId
#
# Expected credentials looks a little like this:
# serviceId: 84920958892    - Random numbers
# servicePassword: XxXXxXXx - Random characters
# spaceId: 984348           - Random numbers
#
# 1. Visit https://www.sfr.fr/
#
# 2. Url will look like this
#    https://www.dmc.sfr-sh.fr/DmcWS/1.5.8/JsonService/<apiGroup>/<apicall>

import json

import requests

from ..common import NotifyType
from ..locale import gettext_lazy as _
from ..url import PrivacyMode
from ..utils.parse import is_phone_no, parse_phone_no
from .base import NotifyBase


class NotifySFR(NotifyBase):
    """A wrapper for SFR French Telecom DMC API."""

    # The default descriptive name associated with the Notification
    service_name = _("Société Française du Radiotéléphone")

    # The services URL
    service_url = "https://www.sfr.fr/"

    # The default protocol
    protocol = "sfr"

    # A URL that takes you to the setup/help of the specific protocol
    setup_url = "https://github.com/caronc/apprise/wiki/Notify_sfr"

    # SFR api
    notify_url = (
        "https://www.dmc.sfr-sh.fr/DmcWS/1.5.8/JsonService/"
        "MessagesUnitairesWS/addSingleCall"  # this is the actual api call
    )

    # The maximum length of the body
    body_maxlen = 160

    # A title can not be used for SMS Messages.  Setting this to zero will
    # cause any title (if defined) to get placed into the message body.
    title_maxlen = 0

    # Define object templates
    templates = ("{schema}://{user}:{password}@{space_id}/{targets}",)

    # Define our tokens
    template_tokens = dict(
        NotifyBase.template_tokens,
        **{
            "user": {
                "name": _("Service ID"),
                "type": "string",
                "required": True,
            },
            "password": {
                "name": _("Service Password"),
                "type": "string",
                "private": True,
                "required": True,
            },
            "space_id": {
                "name": _("Space ID"),
                "type": "string",
                "private": True,
                "required": True,
            },
            "target": {
                "name": _("Recipient Phone Number"),
                "type": "string",
                "regex": (r"^\+?[0-9\s)(+-]+$", "i"),
                "map_to": "targets",
            },
            "targets": {
                "name": _("Targets"),
                "type": "list:string",
                "required": True,
            },
        },
    )

    # Define our template arguments
    template_args = dict(
        NotifyBase.template_args,
        **{
            "lang": {
                "name": _("Language"),
                "type": "string",
                "default": "fr_FR",
                "required": True,
            },
            "sender": {
                "name": _("Sender Name"),
                "type": "string",
                "required": True,
                "default": "",
            },
            "from": {"alias_of": "sender"},
            "media": {
                "name": _("Media Type"),
                "type": "string",
                "required": True,
                "default": "SMSUnicode",
                "values": ["SMS", "SMSLong", "SMSUnicode", "SMSUnicodeLong"],
            },
            "timeout": {
                "name": _("Timeout"),
                "type": "int",
                "default": 2880,
                "required": False,
            },
            "voice": {
                "name": _("TTS Voice"),
                "type": "string",
                "default": "claire08s",
                "values": ["claire08s", "laura8k"],
                "required": False,
            },
            "to": {
                "alias_of": "targets",
            },
        },
    )

    def __init__(
        self,
        space_id=None,
        targets=None,
        lang=None,
        sender=None,
        media=None,
        timeout=None,
        voice=None,
        **kwargs,
    ):
        """Initialize SFR Object."""
        super().__init__(**kwargs)

        if not (self.user and self.password):
            msg = (
                "A SFR user (serviceId) and password (servicePassword) "
                "combination was not provided."
            )
            self.logger.warning(msg)
            raise TypeError(msg)

        self.space_id = space_id
        if not self.space_id:
            msg = "A SFR Space ID is required."
            self.logger.warning(msg)
            raise TypeError(msg)

        self.voice = voice if voice else self.template_args["voice"]["default"]
        self.lang = lang if lang else self.template_args["lang"]["default"]
        self.media = media if media else self.template_args["media"]["default"]
        self.sender = (
            sender if sender else self.template_args["sender"]["default"]
        )

        # Set our Time to Live Flag
        self.timeout = self.template_args["timeout"]["default"]
        try:
            self.timeout = int(timeout)

        except (ValueError, TypeError):
            # set default timeout
            self.timeout = 2880
            pass

        # Parse our targets
        self.targets = []

        for target in parse_phone_no(targets):
            # Validate targets and drop bad ones:
            result = is_phone_no(target)
            if not result:
                self.logger.warning(
                    f"Dropped invalid phone # ({target}) specified.",
                )
                continue

            # store valid phone number
            self.targets.append(result["full"])

        if not self.targets:
            msg = (
                "No receiver phone number has been provided. Please "
                "provide as least one valid phone number."
            )
            self.logger.warning(msg)
            raise TypeError(msg)

        return

    def send(self, body, title="", notify_type=NotifyType.INFO, **kwargs):
        """Perform the SFR notification."""

        # error tracking (used for function return)
        has_error = False

        # Create a copy of the targets list
        targets = list(self.targets)

        # Construct the authentication JSON
        auth_payload = json.dumps({
            "serviceId": self.user,
            "servicePassword": self.password,
            "spaceId": self.space_id,
            "lang": self.lang,
        })

        base_payload = {
            # Can be 'SMS', 'SMSLong', 'SMSUnicode', or 'SMSUnicodeLong'
            "media": self.media,
            # Content of the message
            "textMsg": body,
            # Receiver's phone number (set below)
            "to": None,
            # Optional, default to ''
            "from": self.sender,
            # Optional, default 2880 minutes
            "timeout": self.timeout,
            # Optional, default to French voice
            "ttsVoice": self.voice,
        }

        while len(targets):
            # Get our target to notify
            target = targets.pop(0)

            # Prepare our target phone no
            base_payload["to"] = target

            # Always call throttle before any remote server i/o is made
            self.throttle()

            # Finalize our payload
            payload = {
                "authenticate": auth_payload,
                "messageUnitaire": json.dumps(base_payload, ensure_ascii=True),
            }

            # Some Debug Logging
            self.logger.debug(
                "SFR POST URL:"
                f" {self.notify_url} (cert_verify={self.verify_certificate})"
            )
            self.logger.debug(f"SFR Payload: {payload}")

            try:
                r = requests.post(
                    self.notify_url,
                    params=payload,
                    verify=self.verify_certificate,
                    timeout=self.request_timeout,
                )

                try:
                    content = json.loads(r.content)

                except (AttributeError, TypeError, ValueError):
                    # ValueError = r.content is Unparsable
                    # TypeError = r.content is None
                    # AttributeError = r is None
                    content = {}

                # Check if the request was successfull
                if r.status_code not in (
                    requests.codes.ok,
                    requests.codes.no_content,
                ):
                    # We had a problem
                    status_str = NotifySFR.http_response_code_lookup(
                        r.status_code
                    )

                    self.logger.warning(
                        "Failed to send SFR notification to {}: "
                        "{}{}error={}.".format(
                            target,
                            status_str,
                            ", " if status_str else "",
                            r.status_code,
                        )
                    )

                    self.logger.debug(f"Response Details:\r\n{r.content}")

                    # Mark our failure
                    has_error = True
                    continue

                # SFR returns a code 200 even if the authentication fails
                # It then indicates in the content['success'] field the
                # Actual state of the transaction
                if not content.get("success", False):
                    self.logger.warning(
                        "SFR Notification to {} was not sent by the server: "
                        "server_error={}, fatal={}.".format(
                            target,
                            content.get("errorCode", "UNKNOWN"),
                            content.get("fatal", "True"),
                        )
                    )

                    # Mark our failure
                    has_error = True
                    continue

                self.logger.info(f"Sent SFR notification to {target}.")

            except requests.RequestException as e:
                self.logger.warning(
                    f"A Connection error occurred sending SFR:{target} "
                    "notification."
                )
                self.logger.debug(f"Socket Exception: {e!s}")

                # Mark our failure
                has_error = True
                continue

        return not has_error

    @property
    def url_identifier(self):
        """Returns all of the identifiers that make this URL unique from
        another simliar one.

        Targets or end points should never be identified here.
        """
        return (
            self.secure_protocol if self.secure else self.protocol,
            self.user,
            self.password,
            self.space_id,
        )

    def url(self, privacy=False, *args, **kwargs):
        """Returns the URL built dynamically based on specified arguments."""
        # Define any URL parameters
        params = {
            "from": self.sender,
            "timeout": str(self.timeout),
            "voice": self.voice,
            "lang": self.lang,
            "media": self.media,
        }

        # Extend our parameters
        params.update(self.url_parameters(privacy=privacy, *args, **kwargs))

        return "{schema}://{user}:{password}@{sid}/{targets}?{params}".format(
            schema=self.secure_protocol if self.secure else self.protocol,
            user=self.user,
            password=self.pprint(
                self.password,
                privacy,
                mode=PrivacyMode.Secret,
                safe="",
            ),
            sid=self.pprint(self.space_id, privacy, safe=""),
            targets="/".join(
                [NotifySFR.quote(x, safe="") for x in self.targets]
            ),
            params=self.urlencode(params),
        )

    def __len__(self):
        """Returns the number of targets associated with this notification."""
        return len(self.targets)

    @staticmethod
    def parse_url(url):
        """Parse the URL and return arguments required to initialize this
        plugin."""
        # NotifyBase.parse_url() will make the initial parsing of your string
        # very easy to use. It will tokenize the entire URL for you.  The
        # tokens are then passed into your __init__() function you defined to
        # generate you're object

        results = NotifyBase.parse_url(url, verify_host=False)

        if not results:
            # We're done early as we couldn't load the results
            return results

        # Extract user and password
        results["space_id"] = results.get("host")
        results["targets"] = NotifySFR.split_path(results["fullpath"])

        # Extract additional parameters
        qsd = results.get("qsd", {})
        results["sender"] = NotifySFR.unquote(
            qsd.get("sender", qsd.get("from"))
        )
        results["timeout"] = NotifySFR.unquote(qsd.get("timeout"))
        results["voice"] = NotifySFR.unquote(qsd.get("voice"))
        results["lang"] = NotifySFR.unquote(qsd.get("lang"))
        results["media"] = NotifySFR.unquote(qsd.get("media"))

        # Support the 'to' variable so that we can support rooms this way too
        # The 'to' makes it easier to use yaml configuration
        if "to" in results["qsd"] and len(results["qsd"]["to"]):
            results["targets"] += NotifySFR.parse_phone_no(
                results["qsd"]["to"]
            )

        return results



# BSD 2-Clause License
#
# Apprise - Push Notification Library.
# Copyright (c) 2025, Chris Caron <lead2gold@gmail.com>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

#
# API Docs: https://developers.line.biz/en/reference/messaging-api/

from json import dumps
import re

import requests

from ..common import NotifyImageSize, NotifyType
from ..locale import gettext_lazy as _
from ..url import PrivacyMode
from ..utils.parse import parse_bool, parse_list, validate_regex
from .base import NotifyBase

# Used to break path apart into list of streams
TARGET_LIST_DELIM = re.compile(r"[ \t\r\n,#\\/]+")


class NotifyLine(NotifyBase):
    """A wrapper for Line Notifications."""

    # The default descriptive name associated with the Notification
    service_name = "Line"

    # The services URL
    service_url = "https://line.me/"

    # Secure Protocol
    secure_protocol = "line"

    # The URL refererenced for remote Notifications
    notify_url = "https://api.line.me/v2/bot/message/push"

    # A URL that takes you to the setup/help of the specific protocol
    setup_url = "https://github.com/caronc/apprise/wiki/Notify_line"

    # We don't support titles for Line notifications
    title_maxlen = 0

    # Maximum body length is 5000
    body_maxlen = 5000

    # Allows the user to specify the NotifyImageSize object; this is supported
    # through the webhook
    image_size = NotifyImageSize.XY_128

    # Define object templates
    templates = ("{schema}://{token}/{targets}",)

    # Define our template tokens
    template_tokens = dict(
        NotifyBase.template_tokens,
        **{
            "token": {
                "name": _("Access Token"),
                "type": "string",
                "private": True,
                "required": True,
            },
            "target_user": {
                "name": _("Target User"),
                "type": "string",
                "map_to": "targets",
            },
            "targets": {
                "name": _("Targets"),
                "type": "list:string",
                "required": True,
            },
        },
    )

    # Define our template arguments
    template_args = dict(
        NotifyBase.template_args,
        **{
            "to": {
                "alias_of": "targets",
            },
            "image": {
                "name": _("Include Image"),
                "type": "bool",
                "default": True,
                "map_to": "include_image",
            },
        },
    )

    def __init__(self, token, targets=None, include_image=True, **kwargs):
        """Initialize Line Object."""
        super().__init__(**kwargs)

        # Long-Lived Access token (generated from User Profile)
        self.token = validate_regex(token)
        if not self.token:
            msg = f"An invalid Access Token ({token}) was specified."
            self.logger.warning(msg)
            raise TypeError(msg)

        # Display our Apprise Image
        self.include_image = include_image

        # Set up our targets
        self.targets = parse_list(targets)

        # A dictionary of cached users
        self.__cached_users = {}

        return

    def send(self, body, title="", notify_type=NotifyType.INFO, **kwargs):
        """Send our Line Notification."""

        if len(self.targets) == 0:
            # There were no services to notify
            self.logger.warning("There were no Line targets to notify.")
            return False

        # error tracking (used for function return)
        has_error = False

        # Prepare our headers
        headers = {
            "User-Agent": self.app_id,
            "Content-Type": "application/json",
            "Authorization": f"Bearer {self.token}",
        }

        # Prepare our persistent_notification.create payload
        payload = {
            "to": None,
            "messages": [{
                "type": "text",
                "text": body,
                "sender": {
                    "name": self.app_id,
                },
            }],
        }

        # Acquire our image url if configured to do so
        image_url = (
            None if not self.include_image else self.image_url(notify_type)
        )

        if image_url:
            payload["messages"][0]["sender"]["iconUrl"] = image_url

        # Create a copy of the target list
        targets = list(self.targets)
        while len(targets):
            target = targets.pop(0)

            payload["to"] = target

            self.logger.debug(
                "Line POST URL:"
                f" {self.notify_url} (cert_verify={self.verify_certificate!r})"
            )
            self.logger.debug(f"Line Payload: {payload!s}")

            # Always call throttle before any remote server i/o is made
            self.throttle()
            try:
                r = requests.post(
                    self.notify_url,
                    data=dumps(payload),
                    headers=headers,
                    verify=self.verify_certificate,
                    timeout=self.request_timeout,
                )
                if r.status_code != requests.codes.ok:
                    # We had a problem
                    status_str = NotifyLine.http_response_code_lookup(
                        r.status_code
                    )

                    self.logger.warning(
                        "Failed to send Line notification to {}: "
                        "{}{}error={}.".format(
                            target,
                            status_str,
                            ", " if status_str else "",
                            r.status_code,
                        )
                    )

                    self.logger.debug(f"Response Details:\r\n{r.content}")

                    # Mark our failure
                    has_error = True
                    continue

                else:
                    self.logger.info(f"Sent Line notification to {target}.")

            except requests.RequestException as e:
                self.logger.warning(
                    "A Connection error occurred sending Line "
                    f"notification to {target}."
                )
                self.logger.debug(f"Socket Exception: {e!s}")

                # Mark our failure
                has_error = True
                continue

        return not has_error

    @property
    def url_identifier(self):
        """Returns all of the identifiers that make this URL unique from
        another simliar one.

        Targets or end points should never be identified here.
        """
        return (self.secure_protocol, self.token)

    def url(self, privacy=False, *args, **kwargs):
        """Returns the URL built dynamically based on specified arguments."""

        # Define any URL parameters
        params = {
            "image": "yes" if self.include_image else "no",
        }

        # Extend our parameters
        params.update(self.url_parameters(privacy=privacy, *args, **kwargs))

        return "{schema}://{token}/{targets}?{params}".format(
            schema=self.secure_protocol,
            # never encode hostname since we're expecting it to be a valid one
            token=self.pprint(
                self.token, privacy, mode=PrivacyMode.Secret, safe=""
            ),
            targets="/".join(
                [self.pprint(x, privacy, safe="") for x in self.targets]
            ),
            params=NotifyLine.urlencode(params),
        )

    def __len__(self):
        """Returns the number of targets associated with this notification."""
        return len(self.targets)

    @staticmethod
    def parse_url(url):
        """Parses the URL and returns enough arguments that can allow us to re-
        instantiate this object."""

        results = NotifyBase.parse_url(url, verify_host=False)
        if not results:
            # We're done early as we couldn't load the results
            return results

        # Get unquoted entries
        results["targets"] = NotifyLine.split_path(results["fullpath"])

        # The 'token' makes it easier to use yaml configuration
        if "token" in results["qsd"] and len(results["qsd"]["token"]):
            results["token"] = NotifyLine.unquote(results["qsd"]["token"])
        else:
            results["token"] = NotifyLine.unquote(results["host"])

            # Line Long Lived Tokens included forward slashes in them.
            # As a result we need to parse further into our path and look
            # for the entry that ends in an equal symbol.
            if not results["token"].endswith("="):
                for index, entry in enumerate(
                    list(results["targets"]), start=1
                ):
                    if entry.endswith("="):
                        # Found
                        results["token"] += "/" + "/".join(
                            results["targets"][0:index]
                        )
                        results["targets"] = results["targets"][index:]
                        break

        # Include images with our message
        results["include_image"] = parse_bool(
            results["qsd"].get("image", True)
        )

        # Support the 'to' variable so that we can support rooms this way too
        # The 'to' makes it easier to use yaml configuration
        if "to" in results["qsd"] and len(results["qsd"]["to"]):
            results["targets"] += list(
                filter(
                    bool,
                    TARGET_LIST_DELIM.split(
                        NotifyLine.unquote(results["qsd"]["to"])
                    ),
                )
            )

        return results



# BSD 2-Clause License
#
# Apprise - Push Notification Library.
# Copyright (c) 2025, Chris Caron <lead2gold@gmail.com>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
import re

from .parse import is_hostname, parse_url


def cwe312_word(word, force=False, advanced=True, threshold=5):
    """This function was written to help mask secure/private information that
    may or may not be found within Apprise. The idea is to provide a
    presentable word response that the user who prepared it would understand,
    yet not reveal any private information for any potential intruder.

    For more detail see CWE-312 @
       https://cwe.mitre.org/data/definitions/312.html

    The `force` is an optional argument used to keep the string formatting
    consistent and in one place. If set, the content passed in is presumed
    to be containing secret information and will be updated accordingly.

    If advanced is set to `True` then content is additionally checked for
    upper/lower/ascii/numerical variances. If an obscurity threshold is
    reached, then content is considered secret
    """

    class Variance:
        """A Simple List of Possible Character Variances."""

        # An Upper Case Character (ABCDEF... etc)
        ALPHA_UPPER = "+"
        # An Lower Case Character (abcdef... etc)
        ALPHA_LOWER = "-"
        # A Special Character ($%^;... etc)
        SPECIAL = "s"
        # A Numerical Character (1234... etc)
        NUMERIC = "n"

    if not (isinstance(word, str) and word.strip()):
        # not a password if it's not something we even support
        return word

    # Formatting
    word = word.strip()
    if force:
        # We're forcing the representation to be a secret
        # We do this for consistency
        return f"{word[0:1]}...{word[-1:]}"

    elif len(word) > 1 and not is_hostname(
        word, ipv4=True, ipv6=True, underscore=False
    ):
        # Verify if it is a hostname or not
        return f"{word[0:1]}...{word[-1:]}"

    elif len(word) >= 16:
        # an IP will be 15 characters so we don't want to use a smaller
        # value then 16 (e.g 101.102.103.104)
        # we can assume very long words are passwords otherwise
        return f"{word[0:1]}...{word[-1:]}"

    if advanced:
        #
        # Mark word a secret based on it's obscurity
        #

        # Our variances will increase depending on these variables:
        last_variance = None
        obscurity = 0

        for c in word:
            # Detect our variance
            if c.isdigit():
                variance = Variance.NUMERIC
            elif c.isalpha() and c.isupper():
                variance = Variance.ALPHA_UPPER
            elif c.isalpha() and c.islower():
                variance = Variance.ALPHA_LOWER
            else:
                variance = Variance.SPECIAL

            if last_variance != variance or variance == Variance.SPECIAL:
                obscurity += 1

                if obscurity >= threshold:
                    return f"{word[0:1]}...{word[-1:]}"

            last_variance = variance

    # Otherwise we're good; return our word
    return word


def cwe312_url(url):
    """This function was written to help mask secure/private information that
    may or may not be found on an Apprise URL. The idea is to not disrupt the
    structure of the previous URL too much, yet still protect the users private
    information from being logged directly to screen.

    For more detail see CWE-312 @
    https://cwe.mitre.org/data/definitions/312.html

    For example, consider the URL: http://user:password@localhost/

    When passed into this function, the return value would be:
    http://user:****@localhost/

    Since apprise allows you to put private information everywhere in it's
    custom URLs, it uses this function to manipulate the content before
    returning to any kind of logger.

    The idea is that the URL can still be interpreted by the person who
    constructed them, but not to an intruder.
    """
    # Parse our URL
    results = parse_url(url)
    if not results:
        # Nothing was returned (invalid data was fed in); return our
        # information as it was fed to us (without changing it)
        return url

    # Update our URL with values
    results["password"] = cwe312_word(results["password"], force=True)
    if not results["schema"].startswith("http"):
        results["user"] = cwe312_word(results["user"])
        results["host"] = cwe312_word(results["host"])

    else:
        results["host"] = cwe312_word(results["host"], advanced=False)
        results["user"] = cwe312_word(results["user"], advanced=False)

    # Apply our full path scan in all cases
    results["fullpath"] = (
        "/"
        + "/".join([
            cwe312_word(x)
            for x in re.split(r"[\\/]+", results["fullpath"].lstrip("/"))
        ])
        if results["fullpath"]
        else ""
    )

    #
    # Now re-assemble our URL for display purposes
    #

    # Determine Authentication
    auth = ""
    if results["user"] and results["password"]:
        auth = "{user}:{password}@".format(
            user=results["user"],
            password=results["password"],
        )
    elif results["user"]:
        auth = "{user}@".format(
            user=results["user"],
        )

    params = ""
    if results["qsd"]:
        params = "?{}".format(
            "&".join([
                "{}={}".format(
                    k,
                    cwe312_word(
                        v,
                        force=(
                            k
                            in (
                                "password",
                                "secret",
                                "pass",
                                "token",
                                "key",
                                "id",
                                "apikey",
                                "to",
                            )
                        ),
                    ),
                )
                for k, v in results["qsd"].items()
            ])
        )

    return "{schema}://{auth}{hostname}{port}{fullpath}{params}".format(
        schema=results["schema"],
        auth=auth,
        # never encode hostname since we're expecting it to be a valid one
        hostname=results["host"],
        port="" if not results["port"] else ":{}".format(results["port"]),
        fullpath=results["fullpath"] if results["fullpath"] else "",
        params=params,
    )



# BSD 2-Clause License
#
# Apprise - Push Notification Library.
# Copyright (c) 2025, Chris Caron <lead2gold@gmail.com>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.


import re

from .common import SecureMailMode, WebBaseLogin

# To attempt to make this script stupid proof, if we detect an email address
# that is part of the this table, we can pre-use a lot more defaults if they
# aren't otherwise specified on the users input.
EMAIL_TEMPLATES = (
    # Google GMail
    (
        "Google Mail",
        re.compile(
            r"^((?P<label>[^+]+)\+)?(?P<id>[^@]+)@" r"(?P<domain>gmail\.com)$",
            re.I,
        ),
        {
            "port": 587,
            "smtp_host": "smtp.gmail.com",
            "secure": True,
            "secure_mode": SecureMailMode.STARTTLS,
            "login_type": (WebBaseLogin.EMAIL,),
        },
    ),
    # Yandex
    (
        "Yandex",
        re.compile(
            r"^((?P<label>[^+]+)\+)?(?P<id>[^@]+)@"
            r"(?P<domain>yandex\.(com|ru|ua|by|kz|uz|tr|fr))$",
            re.I,
        ),
        {
            "port": 465,
            "smtp_host": "smtp.yandex.ru",
            "secure": True,
            "secure_mode": SecureMailMode.SSL,
            "login_type": (WebBaseLogin.USERID,),
        },
    ),
    # Microsoft Hotmail
    (
        "Microsoft Hotmail",
        re.compile(
            r"^((?P<label>[^+]+)\+)?(?P<id>[^@]+)@"
            r"(?P<domain>(hotmail|live)\.com(\.au)?)$",
            re.I,
        ),
        {
            "port": 587,
            "smtp_host": "smtp-mail.outlook.com",
            "secure": True,
            "secure_mode": SecureMailMode.STARTTLS,
            "login_type": (WebBaseLogin.EMAIL,),
        },
    ),
    # Microsoft Outlook
    (
        "Microsoft Outlook",
        re.compile(
            r"^((?P<label>[^+]+)\+)?(?P<id>[^@]+)@"
            r"(?P<domain>(smtp\.)?outlook\.com(\.au)?)$",
            re.I,
        ),
        {
            "port": 587,
            "smtp_host": "smtp.outlook.com",
            "secure": True,
            "secure_mode": SecureMailMode.STARTTLS,
            "login_type": (WebBaseLogin.EMAIL,),
        },
    ),
    # Microsoft Office 365 (Email Server)
    # You must specify an authenticated sender address in the from= settings
    # and a valid email in the to= to deliver your emails to
    (
        "Microsoft Office 365",
        re.compile(r"^[^@]+@(?P<domain>(smtp\.)?office365\.com)$", re.I),
        {
            "port": 587,
            "smtp_host": "smtp.office365.com",
            "secure": True,
            "secure_mode": SecureMailMode.STARTTLS,
        },
    ),
    # Yahoo Mail
    (
        "Yahoo Mail",
        re.compile(
            r"^((?P<label>[^+]+)\+)?(?P<id>[^@]+)@"
            r"(?P<domain>yahoo\.(ca|com))$",
            re.I,
        ),
        {
            "port": 465,
            "smtp_host": "smtp.mail.yahoo.com",
            "secure": True,
            "secure_mode": SecureMailMode.STARTTLS,
            "login_type": (WebBaseLogin.EMAIL,),
        },
    ),
    # Fast Mail (Series 1)
    (
        "Fast Mail",
        re.compile(
            r"^((?P<label>[^+]+)\+)?(?P<id>[^@]+)@"
            r"(?P<domain>fastmail\.(com|cn|co\.uk|com\.au|de|es|fm|fr|im|"
            r"in|jp|mx|net|nl|org|se|to|tw|uk|us))$",
            re.I,
        ),
        {
            "port": 465,
            "smtp_host": "smtp.fastmail.com",
            "secure": True,
            "secure_mode": SecureMailMode.SSL,
            "login_type": (WebBaseLogin.EMAIL,),
        },
    ),
    # Fast Mail (Series 2)
    (
        "Fast Mail Extended Addresses",
        re.compile(
            r"^((?P<label>[^+]+)\+)?(?P<id>[^@]+)@"
            r"(?P<domain>123mail\.org|airpost\.net|eml\.cc|fmail\.co\.uk|"
            r"fmgirl\.com|fmguy\.com|mailbolt\.com|mailcan\.com|"
            r"mailhaven\.com|mailmight\.com|ml1\.net|mm\.st|myfastmail\.com|"
            r"proinbox\.com|promessage\.com|rushpost\.com|sent\.(as|at|com)|"
            r"speedymail\.org|warpmail\.net|xsmail\.com|150mail\.com|"
            r"150ml\.com|16mail\.com|2-mail\.com|4email\.net|50mail\.com|"
            r"allmail\.net|bestmail\.us|cluemail\.com|elitemail\.org|"
            r"emailcorner\.net|emailengine\.(net|org)|emailgroups\.net|"
            r"emailplus\.org|emailuser\.net|f-m\.fm|fast-email\.com|"
            r"fast-mail\.org|fastem\.com|fastemail\.us|fastemailer\.com|"
            r"fastest\.cc|fastimap\.com|fastmailbox\.net|fastmessaging\.com|"
            r"fea\.st|fmailbox\.com|ftml\.net|h-mail\.us|hailmail\.net|"
            r"imap-mail\.com|imap\.cc|imapmail\.org|inoutbox\.com|"
            r"internet-e-mail\.com|internet-mail\.org|internetemails\.net|"
            r"internetmailing\.net|jetemail\.net|justemail\.net|"
            r"letterboxes\.org|mail-central\.com|mail-page\.com|"
            r"mailandftp\.com|mailas\.com|mailc\.net|mailforce\.net|"
            r"mailftp\.com|mailingaddress\.org|mailite\.com|mailnew\.com|"
            r"mailsent\.net|mailservice\.ms|mailup\.net|mailworks\.org|"
            r"mymacmail\.com|nospammail\.net|ownmail\.net|petml\.com|"
            r"postinbox\.com|postpro\.net|realemail\.net|reallyfast\.biz|"
            r"reallyfast\.info|speedpost\.net|ssl-mail\.com|swift-mail\.com|"
            r"the-fastest\.net|the-quickest\.com|theinternetemail\.com|"
            r"veryfast\.biz|veryspeedy\.net|yepmail\.net)$",
            re.I,
        ),
        {
            "port": 465,
            "smtp_host": "smtp.fastmail.com",
            "secure": True,
            "secure_mode": SecureMailMode.SSL,
            "login_type": (WebBaseLogin.EMAIL,),
        },
    ),
    # Zoho Mail (Free)
    (
        "Zoho Mail",
        re.compile(
            r"^((?P<label>[^+]+)\+)?(?P<id>[^@]+)@"
            r"(?P<domain>zoho(mail)?\.com)$",
            re.I,
        ),
        {
            "port": 587,
            "smtp_host": "smtp.zoho.com",
            "secure": True,
            "secure_mode": SecureMailMode.STARTTLS,
            "login_type": (WebBaseLogin.EMAIL,),
        },
    ),
    # SendGrid (Email Server)
    # You must specify an authenticated sender address in the from= settings
    # and a valid email in the to= to deliver your emails to
    (
        "SendGrid",
        re.compile(
            r"^((?P<label>[^+]+)\+)?(?P<id>[^@]+)@"
            r"(?P<domain>(\.smtp)?sendgrid\.(com|net))$",
            re.I,
        ),
        {
            "port": 465,
            "smtp_host": "smtp.sendgrid.net",
            "secure": True,
            "secure_mode": SecureMailMode.SSL,
            "login_type": (WebBaseLogin.USERID,),
        },
    ),
    # 163.com
    (
        "163.com",
        re.compile(
            r"^((?P<label>[^+]+)\+)?(?P<id>[^@]+)@" r"(?P<domain>163\.com)$",
            re.I,
        ),
        {
            "port": 465,
            "smtp_host": "smtp.163.com",
            "secure": True,
            "secure_mode": SecureMailMode.SSL,
            "login_type": (WebBaseLogin.EMAIL,),
        },
    ),
    # Foxmail.com
    (
        "Foxmail.com",
        re.compile(
            r"^((?P<label>[^+]+)\+)?(?P<id>[^@]+)@"
            r"(?P<domain>(foxmail|qq)\.com)$",
            re.I,
        ),
        {
            "port": 587,
            "smtp_host": "smtp.qq.com",
            "secure": True,
            "secure_mode": SecureMailMode.STARTTLS,
            "login_type": (WebBaseLogin.EMAIL,),
        },
    ),
    # Comcast.net
    (
        "Comcast.net",
        re.compile(
            r"^((?P<label>[^+]+)\+)?(?P<id>[^@]+)@"
            r"(?P<domain>(comcast)\.net)$",
            re.I,
        ),
        {
            "port": 465,
            "smtp_host": "smtp.comcast.net",
            "secure": True,
            "secure_mode": SecureMailMode.SSL,
            "login_type": (WebBaseLogin.EMAIL,),
        },
    ),
    # Localhost handling
    (
        "Local Mail Server",
        re.compile(
            r"^(((?P<label>[^+]+)\+)?(?P<id>[^@]+)@)?"
            r"(?P<domain>localhost(\.localdomain)?)$",
            re.I,
        ),
        {
            # Provide a default user if one isn't provided
            "from_user": "root",
            "smtp_host": None,
        },
    ),
    # Catch All
    (
        "Custom",
        re.compile(
            r"^((?P<label>[^+]+)\+)?(?P<id>[^@]+)@" r"(?P<domain>.+)$", re.I
        ),
        {
            # Setting smtp_host to None is a way of
            # auto-detecting it based on other parameters
            # specified.  There is no reason to ever modify
            # this Catch All
            "smtp_host": None,
        },
    ),
)



# BSD 2-Clause License
#
# Apprise - Push Notification Library.
# Copyright (c) 2025, Chris Caron <lead2gold@gmail.com>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

# Splunk On-Call
# API: https://portal.victorops.com/public/api-docs.html
# Main: https://www.splunk.com/en_us/products/on-call.html
# Routing Keys https://help.victorops.com/knowledge-base/routing-keys/
# Setup: https://help.victorops.com/knowledge-base/rest-endpoint-integration\
#       -guide/


from json import dumps
import re

import requests

from ..common import NOTIFY_TYPES, NotifyType
from ..locale import gettext_lazy as _
from ..utils.parse import validate_regex
from .base import NotifyBase


class SplunkAction:
    """Tracks the actions supported by Apprise Splunk Plugin."""

    # Use mapping (specify :key=arg to over-ride)
    MAP = "map"

    # Creates a timeline event but does not trigger an incident
    INFO = "info"

    # Triggers a warning (possibly causing incident) in all cases
    WARNING = "warning"

    # Triggers an incident in all cases
    CRITICAL = "critical"

    # Acknowldege entity_id provided in all cases
    ACKNOWLEDGE = "acknowledgement"

    # Recovery entity_id provided in all cases
    RECOVERY = "recovery"

    # Resolve (aliase of Recover)
    RESOLVE = "resolve"


# Define our Splunk Actions
SPLUNK_ACTIONS = (
    SplunkAction.MAP,
    SplunkAction.INFO,
    SplunkAction.ACKNOWLEDGE,
    SplunkAction.WARNING,
    SplunkAction.RECOVERY,
    SplunkAction.RESOLVE,
    SplunkAction.CRITICAL,
)


class SplunkMessageType:
    """Defines the supported splunk message types."""

    # Triggers an incident
    CRITICAL = "CRITICAL"

    # May trigger an incident, depending on your settings
    WARNING = "WARNING"

    # Acks an incident
    ACKNOWLEDGEMENT = "ACKNOWLEDGEMENT"

    # Creates a timeline event but does not trigger an incident
    INFO = "INFO"

    # Resolves an incident
    RECOVERY = "RECOVERY"


# Defines our supported message types
SPLUNK_MESSAGE_TYPES = (
    SplunkMessageType.CRITICAL,
    SplunkMessageType.WARNING,
    SplunkMessageType.ACKNOWLEDGEMENT,
    SplunkMessageType.INFO,
    SplunkMessageType.RECOVERY,
)


class NotifySplunk(NotifyBase):
    """A wrapper for Splunk Notifications."""

    # The default descriptive name associated with the Notification
    service_name = _("Splunk On-Call")

    # The services URL
    service_url = "https://www.splunk.com/en_us/products/on-call.html"

    # The default secure protocol
    secure_protocol = ("splunk", "victorops")

    # A URL that takes you to the setup/help of the specific protocol
    setup_url = "https://github.com/caronc/apprise/wiki/Notify_splunk"

    # Notification URL
    notify_url = (
        "https://alert.victorops.com/integrations/generic/20131114/"
        "alert/{apikey}/{routing_key}"
    )

    # Define object templates
    templates = (
        "{schema}://{routing_key}@{apikey}",
        "{schema}://{routing_key}@{apikey}/{entity_id}",
    )

    # The title is not used
    title_maxlen = 60

    # body limit
    body_maxlen = 400

    # Defines our default message mapping
    splunk_message_map = {
        # Creates a timeline event but doesnot trigger an incident
        NotifyType.INFO: SplunkMessageType.INFO,
        # Resolves an incident
        NotifyType.SUCCESS: SplunkMessageType.RECOVERY,
        # May trigger an incident, depending on your settings
        NotifyType.WARNING: SplunkMessageType.WARNING,
        # Triggers an incident
        NotifyType.FAILURE: SplunkMessageType.CRITICAL,
    }

    # Define our tokens; these are the minimum tokens required required to
    # be passed into this function (as arguments). The syntax appends any
    # previously defined in the base package and builds onto them
    template_tokens = dict(
        NotifyBase.template_tokens,
        **{
            "apikey": {
                "name": _("API Key"),
                "type": "string",
                "private": True,
                "required": True,
                "regex": (r"^[A-Z0-9_-]+$", "i"),
            },
            "routing_key": {
                "name": _("Target Routing Key"),
                "type": "string",
                "required": True,
                "regex": (r"^[A-Z0-9_-]+$", "i"),
            },
            "entity_id": {
                # Provide a value such as: "disk space/db01.mycompany.com"
                "name": _("Entity ID"),
                "type": "string",
            },
        },
    )

    # Define our template arguments
    template_args = dict(
        NotifyBase.template_args,
        **{
            "apikey": {
                "alias_of": "apikey",
            },
            "routing_key": {
                "alias_of": "routing_key",
            },
            "route": {
                "alias_of": "routing_key",
            },
            "entity_id": {
                "alias_of": "entity_id",
            },
            "action": {
                "name": _("Action"),
                "type": "choice:string",
                "values": SPLUNK_ACTIONS,
                "default": SPLUNK_ACTIONS[0],
            },
        },
    )

    # Define any kwargs we're using
    template_kwargs = {
        "mapping": {
            "name": _("Action Mapping"),
            "prefix": ":",
        },
    }

    def __init__(
        self,
        apikey,
        routing_key,
        entity_id=None,
        action=None,
        mapping=None,
        **kwargs,
    ):
        """Initialize Splunk Object."""
        super().__init__(**kwargs)

        self.apikey = validate_regex(
            apikey, *self.template_tokens["apikey"]["regex"]
        )
        if not self.apikey:
            msg = f"The Splunk API Key specified ({apikey}) is invalid."
            self.logger.warning(msg)
            raise TypeError(msg)

        self.routing_key = validate_regex(
            routing_key, *self.template_tokens["routing_key"]["regex"]
        )
        if not self.routing_key:
            msg = (
                f"The Splunk Routing Key specified ({routing_key}) is invalid."
            )
            self.logger.warning(msg)
            raise TypeError(msg)

        if not (
            isinstance(entity_id, str) and len(entity_id.strip(" \r\n\t\v/"))
        ):
            # Use routing key
            self.entity_id = f"{self.app_id}/{self.routing_key}"

        else:
            # Assign what was defined:
            self.entity_id = entity_id.strip(" \r\n\t\v/")

        if action and isinstance(action, str):
            self.action = next(
                (a for a in SPLUNK_ACTIONS if a.startswith(action)), None
            )
            if self.action not in SPLUNK_ACTIONS:
                msg = f"The Splunk action specified ({action}) is invalid."
                self.logger.warning(msg)
                raise TypeError(msg)
        else:
            self.action = self.template_args["action"]["default"]

        # Store our mappings
        self.mapping = self.splunk_message_map.copy()
        if mapping and isinstance(mapping, dict):
            for _k, _v in mapping.items():
                # Get our mapping
                k = next((t for t in NOTIFY_TYPES if t.startswith(_k)), None)
                if not k:
                    msg = (
                        f"The Splunk mapping key specified ({_k}) is invalid."
                    )
                    self.logger.warning(msg)
                    raise TypeError(msg)

                _v_upper = _v.upper()
                v = next(
                    (
                        v
                        for v in SPLUNK_MESSAGE_TYPES
                        if v.startswith(_v_upper)
                    ),
                    None,
                )
                if not v:
                    msg = (
                        f"The Splunk mapping value (assigned to {k}) "
                        f"specified ({_v}) is invalid."
                    )
                    self.logger.warning(msg)
                    raise TypeError(msg)

                # Update our mapping
                self.mapping[k] = v

        return

    def send(self, body, title="", notify_type=NotifyType.INFO, **kwargs):
        """Send our notification."""

        # prepare our headers
        headers = {
            "User-Agent": self.app_id,
            "Content-Type": "application/json",
        }

        # Set up our message type
        if self.action == SplunkAction.MAP:
            # Use Mapping
            message_type = self.mapping[notify_type]

        elif self.action == SplunkAction.ACKNOWLEDGE:
            # Always Acknowledge
            message_type = SplunkMessageType.ACKNOWLEDGEMENT

        elif self.action == SplunkAction.INFO:
            # Creates a timeline event but does not trigger an incident
            message_type = SplunkMessageType.INFO

        elif self.action == SplunkAction.CRITICAL:
            # Always create Incident
            message_type = SplunkMessageType.CRITICAL

        elif self.action == SplunkAction.WARNING:
            # Always trigger warning (potentially creating incident)
            message_type = SplunkMessageType.WARNING

        else:  # self.action == SplunkAction.RECOVERY or SplunkAction.RESOLVE
            # Always Recover
            message_type = SplunkMessageType.RECOVERY

        # Prepare our payload
        payload = {
            "entity_id": self.entity_id,
            "message_type": message_type,
            "entity_display_name": title if title else self.app_desc,
            "state_message": body,
            "monitoring_tool": self.app_id,
        }

        notify_url = self.notify_url.format(
            apikey=self.apikey, routing_key=self.routing_key
        )

        self.logger.debug(
            "Splunk GET URL:"
            f" {notify_url} (cert_verify={self.verify_certificate!r})"
        )
        self.logger.debug(f"Splunk Payload: {payload!s}")

        # Always call throttle before any remote server i/o is made
        self.throttle()

        try:
            r = requests.post(
                notify_url,
                data=dumps(payload).encode("utf-8"),
                headers=headers,
                verify=self.verify_certificate,
                timeout=self.request_timeout,
            )
            # Sample Response
            # {
            #   "result" : "success",
            #   "entity_id" : "disk space/db01.mycompany.com"
            # }

            if r.status_code != requests.codes.ok:
                # We had a problem
                status_str = NotifySplunk.http_response_code_lookup(
                    r.status_code
                )

                self.logger.warning(
                    "Failed to send Splunk notification: {}{}error={}.".format(
                        status_str, ", " if status_str else "", r.status_code
                    )
                )

                self.logger.debug(f"Response Details:\r\n{r.content}")

                # Return; we're done
                return False

            else:
                self.logger.info("Sent Splunk notification.")

        except requests.RequestException as e:
            self.logger.warning(
                "A Connection error occurred sending Splunk notification."
            )
            self.logger.debug(f"Socket Exception: {e!s}")

            # Return; we're done
            return False

        return True

    @property
    def url_identifier(self):
        """Returns all of the identifiers that make this URL unique from
        another simliar one.

        Targets or end points should never be identified here.
        """
        return (
            self.secure_protocol[0],
            self.routing_key,
            self.entity_id,
            self.apikey,
        )

    def url(self, privacy=False, *args, **kwargs):
        """Returns the URL built dynamically based on specified arguments."""

        # Define any URL parameters
        params = {
            "action": self.action,
        }

        # Extend our parameters
        params.update(self.url_parameters(privacy=privacy, *args, **kwargs))

        # Append our assignment extra's into our parameters
        params.update({f":{k.value}": v for k, v in self.mapping.items()})

        return "{schema}://{routing_key}@{apikey}/{entity_id}?{params}".format(
            schema=self.secure_protocol[0],
            routing_key=self.routing_key,
            entity_id=(
                "" if self.entity_id == self.routing_key else self.entity_id
            ),
            apikey=self.pprint(self.apikey, privacy, safe=""),
            params=NotifySplunk.urlencode(params),
        )

    @staticmethod
    def parse_url(url):
        """Parses the URL and returns enough arguments that can allow us to re-
        instantiate this object."""

        # parse_url already handles getting the `user` and `password` fields
        # populated.
        results = NotifyBase.parse_url(url, verify_host=False)
        if not results:
            # We're done early as we couldn't load the results
            return results

        # Entity ID
        if "entity_id" in results["qsd"] and len(results["qsd"]["entity_id"]):
            results["entity_id"] = NotifySplunk.unquote(
                results["qsd"]["entity_id"]
            )
        else:
            results["entity_id"] = NotifySplunk.unquote(results["fullpath"])

        # API Key
        if "apikey" in results["qsd"] and len(results["qsd"]["apikey"]):
            results["apikey"] = NotifySplunk.unquote(results["qsd"]["apikey"])

        else:
            results["apikey"] = NotifySplunk.unquote(results["host"])

        # Routing Key
        if "routing_key" in results["qsd"] and len(
            results["qsd"]["routing_key"]
        ):
            results["routing_key"] = NotifySplunk.unquote(
                results["qsd"]["routing_key"]
            )

        elif "route" in results["qsd"] and len(results["qsd"]["route"]):
            results["routing_key"] = NotifySplunk.unquote(
                results["qsd"]["route"]
            )

        else:
            results["routing_key"] = NotifySplunk.unquote(results["user"])

        # Store our action (if defined)
        if "action" in results["qsd"] and len(results["qsd"]["action"]):
            results["action"] = NotifySplunk.unquote(results["qsd"]["action"])

        # store any custom mapping defined
        results["mapping"] = {
            NotifySplunk.unquote(x): NotifySplunk.unquote(y)
            for x, y in results["qsd:"].items()
        }

        return results

    @staticmethod
    def parse_native_url(url):
        """
        Support https://alert.victorops.com/integrations/generic/20131114/ \
                     alert/apikey/routing_key
        """

        result = re.match(
            r"^https?://alert\.victorops\.com/integrations/generic/"
            r"(?P<version>[0-9]+)/alert/(?P<apikey>[0-9a-z_-]+)"
            r"(/(?P<routing_key>[^?/]+))"
            r"(/(?P<entity_id>[^?]+))?/*"
            r"(?P<params>\?.+)?$",
            url,
            re.I,
        )

        if result:
            return NotifySplunk.parse_url(
                "{schema}://{routing_key}@{apikey}/{entity_id}{params}".format(
                    schema=NotifySplunk.secure_protocol[0],
                    apikey=result.group("apikey"),
                    routing_key=result.group("routing_key"),
                    entity_id=(
                        ""
                        if not result.group("entity_id")
                        else result.group("entity_id")
                    ),
                    params=(
                        ""
                        if not result.group("params")
                        else result.group("params")
                    ),
                )
            )

        return None



# BSD 2-Clause License
#
# Apprise - Push Notification Library.
# Copyright (c) 2025, Chris Caron <lead2gold@gmail.com>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

# Free Mobile
#   1. Visit https://mobile.free.fr/

# the URL will look something like this:
#      https://smsapi.free-mobile.fr/sendmsg
#

from json import dumps

import requests

from ..common import NotifyType
from ..locale import gettext_lazy as _
from .base import NotifyBase


class NotifyFreeMobile(NotifyBase):
    """A wrapper for Free-Mobile Notifications."""

    # The default descriptive name associated with the Notification
    service_name = _("Free-Mobile")

    # The services URL
    service_url = "https://mobile.free.fr/"

    # The default secure protocol
    secure_protocol = "freemobile"

    # A URL that takes you to the setup/help of the specific protocol
    setup_url = "https://github.com/caronc/apprise/wiki/Notify_freemobile"

    # Plain Text Notification URL
    notify_url = "https://smsapi.free-mobile.fr/sendmsg"

    # Define object templates
    templates = ("{schema}://{user}@{password}",)

    # The title is not used
    title_maxlen = 0

    # SMS Messages are restricted in size
    body_maxlen = 160

    # Define our tokens; these are the minimum tokens required required to
    # be passed into this function (as arguments). The syntax appends any
    # previously defined in the base package and builds onto them
    template_tokens = dict(
        NotifyBase.template_tokens,
        **{
            "user": {
                "name": _("Username"),
                "type": "string",
                "required": True,
            },
            "password": {
                "name": _("Password"),
                "type": "string",
                "private": True,
                "required": True,
            },
        },
    )

    def __init__(self, **kwargs):
        """Initialize Free Mobile Object."""
        super().__init__(**kwargs)

        if not (self.user and self.password):
            msg = (
                "A FreeMobile user and password combination was not provided."
            )
            self.logger.warning(msg)
            raise TypeError(msg)

        return

    @property
    def url_identifier(self):
        """Returns all of the identifiers that make this URL unique from
        another simliar one.

        Targets or end points should never be identified here.
        """
        return (self.secure_protocol, self.user, self.password)

    def url(self, privacy=False, *args, **kwargs):
        """Returns the URL built dynamically based on specified arguments."""

        # Prepare our parameters
        params = self.url_parameters(privacy=privacy, *args, **kwargs)

        return "{schema}://{user}@{password}/?{params}".format(
            schema=self.secure_protocol,
            user=self.user,
            password=self.pprint(self.password, privacy, safe=""),
            params=NotifyFreeMobile.urlencode(params),
        )

    def send(self, body, title="", notify_type=NotifyType.INFO, **kwargs):
        """Send our notification."""

        # prepare our headers
        headers = {
            "User-Agent": self.app_id,
            "Content-Type": "application/json",
        }

        # Prepare our payload
        payload = {"user": self.user, "pass": self.password, "msg": body}

        self.logger.debug(
            "Free Mobile GET URL:"
            f" {self.notify_url} (cert_verify={self.verify_certificate!r})"
        )
        self.logger.debug(f"Free Mobile Payload: {payload!s}")

        # Always call throttle before any remote server i/o is made
        self.throttle()

        try:
            r = requests.post(
                self.notify_url,
                data=dumps(payload).encode("utf-8"),
                headers=headers,
                verify=self.verify_certificate,
                timeout=self.request_timeout,
            )
            if r.status_code != requests.codes.ok:
                # We had a problem
                status_str = NotifyFreeMobile.http_response_code_lookup(
                    r.status_code
                )

                self.logger.warning(
                    "Failed to send Free Mobile notification: "
                    "{}{}error={}.".format(
                        status_str, ", " if status_str else "", r.status_code
                    )
                )

                self.logger.debug(f"Response Details:\r\n{r.content}")

                # Return; we're done
                return False

            else:
                self.logger.info("Sent Free Mobile notification.")

        except requests.RequestException as e:
            self.logger.warning(
                "A Connection error occurred sending Free Mobile notification."
            )
            self.logger.debug(f"Socket Exception: {e!s}")

            # Return; we're done
            return False

        return True

    @staticmethod
    def parse_url(url):
        """Parses the URL and returns enough arguments that can allow us to re-
        instantiate this object."""

        # parse_url already handles getting the `user` and `password` fields
        # populated.
        results = NotifyBase.parse_url(url, verify_host=False)
        if not results:
            # We're done early as we couldn't load the results
            return results

        # The hostname can act as the password if specified and a password
        # was otherwise not (specified):
        if not results.get("password"):
            results["password"] = NotifyFreeMobile.unquote(results["host"])

        return results



# BSD 2-Clause License
#
# Apprise - Push Notification Library.
# Copyright (c) 2025, Chris Caron <lead2gold@gmail.com>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

# At the time I created this plugin, their website had lots of issues with the
# Firefox Browser.  I fell back to Chrome and had no problems.

# To use this plugin, you need to first access https://teams.webex.com and
# make yourself an account if you don't already have one. You'll want to
# create at least one 'space' before getting the 'incoming webhook'.
#
# Next you'll need to install the 'Incoming webhook' plugin found under
# the 'other' category here: https://apphub.webex.com/integrations/

# These links may not always work as time goes by and websites always
# change, but at the time of creating this plugin this was a direct link
# to it: https://apphub.webex.com/integrations/incoming-webhooks-cisco-systems

# If you're logged in, you'll be able to click on the 'Connect' button. From
# there you'll need to accept the permissions it will ask of you. Give the
# webhook a name such as 'apprise'.
# When you're complete, you will recieve a URL that looks something like this:
# https://api.ciscospark.com/v1/webhooks/incoming/\
#       Y3lzY29zcGkyazovL3VzL1dFQkhPT0sajkkzYWU4fTMtMGE4Yy00
#
# The last part of the URL is all you need to be interested in. Think of this
# url as:
#   https://api.ciscospark.com/v1/webhooks/incoming/{token}
#
# You will need to assemble all of your URLs for this plugin to work as:
#   wxteams://{token}
#
# Resources
# - https://developer.webex.com/docs/api/basics - markdown/post syntax
# - https://developer.cisco.com/ecosystem/webex/apps/\
#       incoming-webhooks-cisco-systems/ - Simple webhook example

from json import dumps
import re

import requests

from ..common import NotifyFormat, NotifyType
from ..locale import gettext_lazy as _
from ..utils.parse import validate_regex
from .base import NotifyBase

# Extend HTTP Error Messages
# Based on: https://developer.webex.com/docs/api/basics/rate-limiting
WEBEX_HTTP_ERROR_MAP = {
    401: "Unauthorized - Invalid Token.",
    415: "Unsuported media specified",
    429: "To many consecutive requests were made.",
    503: "Service is overloaded, try again later",
}


class NotifyWebexTeams(NotifyBase):
    """A wrapper for Webex Teams Notifications."""

    # The default descriptive name associated with the Notification
    service_name = "Cisco Webex Teams"

    # The services URL
    service_url = "https://webex.teams.com/"

    # The default secure protocol
    secure_protocol = ("wxteams", "webex")

    # A URL that takes you to the setup/help of the specific protocol
    setup_url = "https://github.com/caronc/apprise/wiki/Notify_wxteams"

    # Webex Teams uses the http protocol with JSON requests
    notify_url = "https://api.ciscospark.com/v1/webhooks/incoming/"

    # The maximum allowable characters allowed in the body per message
    body_maxlen = 1000

    # We don't support titles for Webex notifications
    title_maxlen = 0

    # Default to markdown; fall back to text
    notify_format = NotifyFormat.MARKDOWN

    # Define object templates
    templates = ("{schema}://{token}",)

    # Define our template tokens
    template_tokens = dict(
        NotifyBase.template_tokens,
        **{
            "token": {
                "name": _("Token"),
                "type": "string",
                "private": True,
                "required": True,
                "regex": (r"^[a-z0-9]{80,160}$", "i"),
            },
        },
    )

    template_args = dict(
        NotifyBase.template_args,
        **{
            "token": {
                "alias_of": "token",
            },
        },
    )

    def __init__(self, token, **kwargs):
        """Initialize Webex Teams Object."""
        super().__init__(**kwargs)

        # The token associated with the account
        self.token = validate_regex(
            token, *self.template_tokens["token"]["regex"]
        )
        if not self.token:
            msg = f"The Webex Teams token specified ({token}) is invalid."
            self.logger.warning(msg)
            raise TypeError(msg)

    def send(self, body, title="", notify_type=NotifyType.INFO, **kwargs):
        """Perform Webex Teams Notification."""

        # Setup our headers
        headers = {
            "User-Agent": self.app_id,
            "Content-Type": "application/json",
        }

        # Prepare our URL
        url = f"{self.notify_url}/{self.token}"

        payload = {
            (
                "markdown"
                if (self.notify_format == NotifyFormat.MARKDOWN)
                else "text"
            ): body,
        }

        self.logger.debug(
            "Webex Teams POST URL:"
            f" {url} (cert_verify={self.verify_certificate!r})"
        )
        self.logger.debug(f"Webex Teams Payload: {payload!s}")

        # Always call throttle before any remote server i/o is made
        self.throttle()
        try:
            r = requests.post(
                url,
                data=dumps(payload),
                headers=headers,
                verify=self.verify_certificate,
                timeout=self.request_timeout,
            )
            if r.status_code not in (
                requests.codes.ok,
                requests.codes.no_content,
            ):
                # We had a problem
                status_str = NotifyWebexTeams.http_response_code_lookup(
                    r.status_code
                )

                self.logger.warning(
                    "Failed to send Webex Teams notification: "
                    "{}{}error={}.".format(
                        status_str, ", " if status_str else "", r.status_code
                    )
                )

                self.logger.debug(f"Response Details:\r\n{r.content}")

                return False

            else:
                self.logger.info("Sent Webex Teams notification.")

        except requests.RequestException as e:
            self.logger.warning(
                "A Connection error occurred sending Webex Teams notification."
            )
            self.logger.debug(f"Socket Exception: {e!s}")

            return False

        return True

    @property
    def url_identifier(self):
        """Returns all of the identifiers that make this URL unique from
        another simliar one.

        Targets or end points should never be identified here.
        """
        return (self.secure_protocol[0], self.token)

    def url(self, privacy=False, *args, **kwargs):
        """Returns the URL built dynamically based on specified arguments."""

        # Our URL parameters
        params = self.url_parameters(privacy=privacy, *args, **kwargs)

        return "{schema}://{token}/?{params}".format(
            schema=self.secure_protocol[0],
            token=self.pprint(self.token, privacy, safe=""),
            params=NotifyWebexTeams.urlencode(params),
        )

    @staticmethod
    def parse_url(url):
        """Parses the URL and returns enough arguments that can allow us to re-
        instantiate this object."""
        results = NotifyBase.parse_url(url, verify_host=False)
        if not results:
            # We're done early as we couldn't load the results
            return results

        # Set our token if found as an argument
        if "token" in results["qsd"] and len(results["qsd"]["token"]):
            results["token"] = \
                NotifyWebexTeams.unquote(results["qsd"]["token"])

        else:
            # The first token is stored in the hostname
            results["token"] = NotifyWebexTeams.unquote(results["host"])

        return results

    @staticmethod
    def parse_native_url(url):
        """
        Support https://api.ciscospark.com/v1/webhooks/incoming/WEBHOOK_TOKEN
        """

        result = re.match(
            r"^https?://(api\.ciscospark\.com|webexapis\.com)"
            r"/v[1-9][0-9]*/webhooks/incoming/"
            r"(?P<webhook_token>[A-Z0-9_-]+)/?"
            r"(?P<params>\?.+)?$",
            url,
            re.I,
        )

        if result:
            return NotifyWebexTeams.parse_url(
                "{schema}://{webhook_token}/{params}".format(
                    schema=NotifyWebexTeams.secure_protocol[0],
                    webhook_token=result.group("webhook_token"),
                    params=(
                        ""
                        if not result.group("params")
                        else result.group("params")
                    ),
                )
            )

        return None



# BSD 2-Clause License
#
# Apprise - Push Notification Library.
# Copyright (c) 2025, Chris Caron <lead2gold@gmail.com>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

# API Details:
# https://docs.microsoft.com/en-us/previous-versions/office/\
#        office-365-api/?redirectedfrom=MSDN

# Information on sending an email:
# https://docs.microsoft.com/en-us/graph/api/user-sendmail\
#       ?view=graph-rest-1.0&tabs=http
#
# Note: One must set up Application Permissions (not Delegated Permissions)
#       - Scopes required: Mail.Send
#       - For Large Attachments: Mail.ReadWrite
#       - For Email Lookups: User.Read.All
#
from datetime import datetime, timedelta
import json
from uuid import uuid4

import requests

from .. import exception
from ..common import NotifyFormat, NotifyType, PersistentStoreMode
from ..locale import gettext_lazy as _
from ..url import PrivacyMode
from ..utils.parse import is_email, parse_emails, validate_regex
from .base import NotifyBase


class NotifyOffice365(NotifyBase):
    """A wrapper for Office 365 Notifications."""

    # The default descriptive name associated with the Notification
    service_name = "Office 365"

    # The services URL
    service_url = "https://office.com/"

    # The default protocol
    secure_protocol = ("azure", "o365")

    # Allow 300 requests per minute.
    # 60/300 = 0.2
    request_rate_per_sec = 0.20

    # A URL that takes you to the setup/help of the specific protocol
    setup_url = "https://github.com/caronc/apprise/wiki/Notify_office365"

    # URL to Microsoft Graph Server
    graph_url = "https://graph.microsoft.com"

    # Authentication URL
    auth_url = "https://login.microsoftonline.com/{tenant}/oauth2/v2.0/token"

    # Support attachments
    attachment_support = True

    # Our default is to no not use persistent storage beyond in-memory
    # reference
    storage_mode = PersistentStoreMode.AUTO

    # the maximum size an attachment can be for it to be allowed to be
    # uploaded inline with the current email going out (one http post)
    # Anything larger than this and a second PUT request is required to
    # the outlook server to post the content through reference.
    # Currently (as of 2025.10.06) this was documented to be 3MB
    outlook_attachment_inline_max = 3145728

    # Use all the direct application permissions you have configured for your
    # app. The endpoint should issue a token for the ones associated with the
    # resource you want to use.
    # see https://docs.microsoft.com/en-us/azure/active-directory/develop/\
    #       v2-permissions-and-consent#the-default-scope
    scope = ".default"

    # Default Notify Format
    notify_format = NotifyFormat.HTML

    # Define object templates
    templates = (
        # Send as user (only supported method)
        "{schema}://{source}/{tenant}/{client_id}/{secret}",
        "{schema}://{source}/{tenant}/{client_id}/{secret}/{targets}",
    )

    # Define our template tokens
    template_tokens = dict(
        NotifyBase.template_tokens,
        **{
            "tenant": {
                "name": _("Tenant Domain"),
                "type": "string",
                "required": True,
                "private": True,
                "regex": (r"^[a-z0-9-]+$", "i"),
            },
            "source": {
                "name": _("Account Email or Object ID"),
                "type": "string",
                "required": True,
            },
            "client_id": {
                "name": _("Client ID"),
                "type": "string",
                "required": True,
                "private": True,
                "regex": (r"^[a-z0-9-]+$", "i"),
            },
            "secret": {
                "name": _("Client Secret"),
                "type": "string",
                "private": True,
                "required": True,
            },
            "target_email": {
                "name": _("Target Email"),
                "type": "string",
                "map_to": "targets",
            },
            "targets": {
                "name": _("Targets"),
                "type": "list:string",
            },
        },
    )

    # Define our template arguments
    template_args = dict(
        NotifyBase.template_args,
        **{
            "to": {
                "alias_of": "targets",
            },
            "cc": {
                "name": _("Carbon Copy"),
                "type": "list:string",
            },
            "bcc": {
                "name": _("Blind Carbon Copy"),
                "type": "list:string",
            },
            "oauth_id": {
                "alias_of": "client_id",
            },
            "oauth_secret": {
                "alias_of": "secret",
            },
        },
    )

    def __init__(
        self,
        tenant,
        client_id,
        secret,
        source=None,
        targets=None,
        cc=None,
        bcc=None,
        **kwargs,
    ):
        """Initialize Office 365 Object."""
        super().__init__(**kwargs)

        # Tenant identifier
        self.tenant = validate_regex(
            tenant, *self.template_tokens["tenant"]["regex"]
        )
        if not self.tenant:
            msg = f"An invalid Office 365 Tenant({tenant}) was specified."
            self.logger.warning(msg)
            raise TypeError(msg)

        # Store our email/ObjectID Source
        self.source = source

        # Client Key (associated with generated OAuth2 Login)
        self.client_id = validate_regex(
            client_id, *self.template_tokens["client_id"]["regex"]
        )
        if not self.client_id:
            msg = (
                "An invalid Office 365 Client OAuth2 ID "
                f"({client_id}) was specified."
            )
            self.logger.warning(msg)
            raise TypeError(msg)

        # Client Secret (associated with generated OAuth2 Login)
        self.secret = validate_regex(secret)
        if not self.secret:
            msg = (
                "An invalid Office 365 Client OAuth2 Secret "
                f"({secret}) was specified."
            )
            self.logger.warning(msg)
            raise TypeError(msg)

        # For tracking our email -> name lookups
        self.names = {}

        # Acquire Carbon Copies
        self.cc = set()

        # Acquire Blind Carbon Copies
        self.bcc = set()

        # Parse our targets
        self.targets = []

        if targets:
            for recipient in parse_emails(targets):
                # Validate recipients (to:) and drop bad ones:
                result = is_email(recipient)
                if result:
                    # Add our email to our target list
                    self.targets.append((
                        result["name"] if result["name"] else False,
                        result["full_email"],
                    ))
                    continue

                self.logger.warning(
                    f"Dropped invalid To email ({recipient}) specified."
                )

        else:
            result = is_email(self.source)
            if not result:
                self.logger.warning("No Target Office 365 Email Detected")

            else:
                # If our target email list is empty we want to add ourselves to
                # it
                self.targets.append((False, self.source))

        # Validate recipients (cc:) and drop bad ones:
        for recipient in parse_emails(cc):
            email = is_email(recipient)
            if email:
                self.cc.add(email["full_email"])

                # Index our name (if one exists)
                self.names[email["full_email"]] = (
                    email["name"] if email["name"] else False
                )
                continue

            self.logger.warning(
                f"Dropped invalid Carbon Copy email ({recipient}) specified.",
            )

        # Validate recipients (bcc:) and drop bad ones:
        for recipient in parse_emails(bcc):
            email = is_email(recipient)
            if email:
                self.bcc.add(email["full_email"])

                # Index our name (if one exists)
                self.names[email["full_email"]] = (
                    email["name"] if email["name"] else False
                )
                continue

            self.logger.warning(
                "Dropped invalid Blind Carbon Copy email "
                f"({recipient}) specified.",
            )

        # Our token is acquired upon a successful login
        self.token = None

        # Presume that our token has expired 'now'
        self.token_expiry = datetime.now()

        # Our email source; we detect this if the source is an ObjectID
        # If it is unknown we set this to None
        # User is the email associated with the account
        self.from_email = self.store.get("from")
        result = is_email(self.source)
        if result:
            self.from_email = result["full_email"]
            self.from_name = result["name"] or self.store.get("name")

        else:
            self.from_name = self.store.get("name")

        return

    def send(
        self,
        body,
        title="",
        notify_type=NotifyType.INFO,
        attach=None,
        **kwargs,
    ):
        """Perform Office 365 Notification."""

        # error tracking (used for function return)
        has_error = False

        if not self.targets:
            # There is no one to email; we're done
            self.logger.warning("There are no Email recipients to notify")
            return False

        if self.from_email is None:
            if not self.authenticate():
                # We could not authenticate ourselves; we're done
                return False

            # Acquire our from_email
            url = f"https://graph.microsoft.com/v1.0/users/{self.source}"
            postokay, response = self._fetch(url=url, method="GET")
            if not postokay:
                self.logger.warning(
                    "Could not acquire From email address; ensure "
                    '"User.Read.All" Application scope is set!'
                )

            else:  # Acquire our from_email (if possible)
                from_email = response.get("mail") or response.get(
                    "userPrincipalName"
                )
                result = is_email(from_email)
                if not result:
                    self.logger.warning(
                        "Could not get From email from the Azure endpoint."
                    )

                    # Prevent re-occuring upstream fetches for info that isn't
                    # there
                    self.from_email = False

                else:
                    # Store our email for future reference
                    self.from_email = result["full_email"]
                    self.store.set("from", result["full_email"])

                    self.from_name = response.get("displayName")
                    if self.from_name:
                        self.store.set("name", self.from_name)

        # Setup our Content Type
        content_type = (
            "HTML" if self.notify_format == NotifyFormat.HTML else "Text"
        )

        # Prepare our payload
        payload = {
            "message": {
                "subject": title,
                "body": {
                    "contentType": content_type,
                    "content": body,
                },
            },
            # Below takes a string (not bool) of either 'true' or 'false'
            "saveToSentItems": "true",
        }

        if self.from_email:
            # Apply from email if it is known
            payload["message"].update({
                "from": {
                    "emailAddress": {
                        "address": self.from_email,
                        "name": self.from_name or self.app_id,
                    }
                },
            })

        # Create a copy of the email list
        emails = list(self.targets)

        # Define our URL to post to
        url = f"{self.graph_url}/v1.0/users/{self.source}/sendMail"

        # Prepare our Draft URL
        draft_url = f"{self.graph_url}/v1.0/users/{self.source}/messages"

        small_attachments = []
        large_attachments = []

        # draft emails
        drafts = []

        if attach and self.attachment_support:
            for no, attachment in enumerate(attach, start=1):
                # Perform some simple error checking
                if not attachment:
                    # We could not access the attachment
                    self.logger.error(
                        "Could not access Office 365 attachment"
                        f" {attachment.url(privacy=True)}."
                    )
                    return False

                if len(attachment) > self.outlook_attachment_inline_max:
                    # Messages larger then xMB need to be uploaded after; a
                    # draft email must be prepared; below is our session
                    large_attachments.append({
                        "obj": attachment,
                        "name": (
                            attachment.name
                            if attachment.name
                            else f"file{no:03}.dat"
                        ),
                    })
                    continue

                try:
                    # Prepare our Attachment in Base64
                    small_attachments.append({
                        "@odata.type": "#microsoft.graph.fileAttachment",
                        # Name of the attachment (as it should appear in email)
                        "name": (
                            attachment.name
                            if attachment.name
                            else f"file{no:03}.dat"
                        ),
                        # MIME type of the attachment
                        "contentType": "attachment.mimetype",
                        # Base64 Content
                        "contentBytes": attachment.base64(),
                    })

                except exception.AppriseException:
                    # We could not access the attachment
                    self.logger.error(
                        "Could not access Office 365 attachment"
                        f" {attachment.url(privacy=True)}."
                    )
                    return False

                self.logger.debug(
                    "Appending Office 365 attachment"
                    f" {attachment.url(privacy=True)}"
                )

        if small_attachments:
            # Store Attachments
            payload["message"]["attachments"] = small_attachments

        while len(emails):
            # authenticate ourselves if we aren't already; but this function
            # also tracks if our token we have is still valid and will
            # re-authenticate ourselves if nessisary.
            if not self.authenticate():
                # We could not authenticate ourselves; we're done
                return False

            # Get our email to notify
            to_name, to_addr = emails.pop(0)

            # Strip target out of cc list if in To or Bcc
            cc = self.cc - self.bcc - {to_addr}

            # Strip target out of bcc list if in To
            bcc = self.bcc - {to_addr}

            # Prepare our email
            payload["message"]["toRecipients"] = [
                {"emailAddress": {"address": to_addr}}
            ]
            if to_name:
                # Apply our To Name
                payload["message"]["toRecipients"][0]["emailAddress"][
                    "name"
                ] = to_name

            self.logger.debug(
                "{}Email To: {}".format(
                    "Draft" if large_attachments else "", to_addr
                )
            )

            if cc:
                # Prepare our CC list
                payload["message"]["ccRecipients"] = []
                for addr in cc:
                    _payload = {"address": addr}
                    if self.names.get(addr):
                        _payload["name"] = self.names[addr]

                    # Store our address in our payload
                    payload["message"]["ccRecipients"].append(
                        {"emailAddress": _payload}
                    )

                self.logger.debug(
                    "{}Email Cc: {}".format(
                        "Draft" if large_attachments else "",
                        ", ".join([
                            "{}{}".format(
                                (
                                    ""
                                    if self.names.get(e)
                                    else f"{self.names[e]}: "
                                ),
                                e,
                            )
                            for e in cc
                        ]),
                    )
                )

            if bcc:
                # Prepare our CC list
                payload["message"]["bccRecipients"] = []
                for addr in bcc:
                    _payload = {"address": addr}
                    if self.names.get(addr):
                        _payload["name"] = self.names[addr]

                    # Store our address in our payload
                    payload["message"]["bccRecipients"].append(
                        {"emailAddress": _payload}
                    )

                self.logger.debug(
                    "{}Email Bcc: {}".format(
                        "Draft" if large_attachments else "",
                        ", ".join([
                            "{}{}".format(
                                (
                                    ""
                                    if self.names.get(e)
                                    else f"{self.names[e]}: "
                                ),
                                e,
                            )
                            for e in bcc
                        ]),
                    )
                )

            # Perform upstream post
            postokay, response = self._fetch(
                url=url if not large_attachments else draft_url,
                payload=payload,
            )

            # Test if we were okay
            if not postokay:
                has_error = True

            elif large_attachments:
                # We have large attachments now to upload and associate with
                # our message. We need to prepare a draft message; acquire
                # the message-id associated with it and then attach the file
                # via this means.

                # Acquire our Draft ID to work with
                message_id = response.get("id")
                if not message_id:
                    self.logger.warning(
                        "Email Draft ID could not be retrieved"
                    )
                    has_error = True
                    continue

                self.logger.debug(f"Email Draft ID: {message_id}")
                # In future, the below could probably be called via async
                has_attach_error = False
                for attachment in large_attachments:
                    if not self.upload_attachment(
                        attachment["obj"], message_id, attachment["name"]
                    ):
                        self.logger.warning(
                            "Could not prepare attachment session for %s",
                            attachment["name"],
                        )

                        has_error = True
                        has_attach_error = True
                        # Take early exit
                        break

                if has_attach_error:
                    continue

                # Send off our draft
                attach_url = (
                    "https://graph.microsoft.com/v1.0/users/"
                    "{}/messages/{}/send"
                )

                attach_url = attach_url.format(
                    self.source,
                    message_id,
                )

                # Trigger our send
                postokay, response = self._fetch(url=url)
                if not postokay:
                    self.logger.warning(
                        "Could not send drafted email id: {} ", message_id
                    )
                    has_error = True
                    continue

        # Memory management
        del small_attachments
        del large_attachments
        del drafts

        return not has_error

    def upload_attachment(self, attachment, message_id, name=None):
        """Uploads an attachment to a session."""

        # Perform some simple error checking
        if not attachment:
            # We could not access the attachment
            self.logger.error(
                "Could not access Office 365 attachment"
                f" {attachment.url(privacy=True)}."
            )
            return False

        # Our Session URL
        url = (
            f"{self.graph_url}/v1.0/users/{self.source}/message/{message_id}"
            + "/attachments/createUploadSession"
        )

        file_size = len(attachment)

        payload = {
            "AttachmentItem": {
                "attachmentType": "file",
                "name": (
                    name
                    if name
                    else (
                        attachment.name
                        if attachment.name
                        else f"{uuid4()!s}.dat"
                    )
                ),
                # MIME type of the attachment
                "contentType": attachment.mimetype,
                "size": file_size,
            }
        }

        if not self.authenticate():
            # We could not authenticate ourselves; we're done
            return False

        # Get our Upload URL
        postokay, response = self._fetch(url, payload)
        if not postokay:
            return False

        upload_url = response.get("uploadUrl")
        if not upload_url:
            return False

        start_byte = 0
        postokay = False
        response = None

        for chunk in attachment.chunk():
            end_byte = start_byte + len(chunk) - 1

            # Define headers for this chunk
            headers = {
                "User-Agent": self.app_id,
                "Content-Length": str(len(chunk)),
                "Content-Range": f"bytes {start_byte}-{end_byte}/{file_size}",
            }

            # Upload the chunk
            postokay, response = self._fetch(
                upload_url,
                chunk,
                headers=headers,
                content_type=None,
                method="PUT",
            )
            if not postokay:
                return False

        # Return our Upload URL
        return postokay

    def authenticate(self):
        """Logs into and acquires us an authentication token to work with."""

        if self.token and self.token_expiry > datetime.now():
            # If we're already authenticated and our token is still valid
            self.logger.debug(f"Already authenticate with token {self.token}")
            return True

        # If we reach here, we've either expired, or we need to authenticate
        # for the first time.

        # Prepare our payload
        payload = {
            "grant_type": "client_credentials",
            "client_id": self.client_id,
            "client_secret": self.secret,
            "scope": f"{self.graph_url}/{self.scope}",
        }

        # Prepare our URL
        url = self.auth_url.format(tenant=self.tenant)

        # A response looks like the following:
        #    {
        #       "token_type": "Bearer",
        #       "expires_in": 3599,
        #       "access_token": "eyJ0eXAiOiJKV1QiLCJhbGciOiJSzI1NiIsInNBXBP..."
        #    }
        #
        # Where expires_in defines the number of seconds the key is valid for
        # before it must be renewed.

        # Alternatively, this could happen too...
        #    {
        #      "error": "invalid_scope",
        #      "error_description": "AADSTS70011: Blah... Blah Blah... Blah",
        #      "error_codes": [
        #        70011
        #      ],
        #      "timestamp": "2020-01-09 02:02:12Z",
        #      "trace_id": "255d1aef-8c98-452f-ac51-23d051240864",
        #      "correlation_id": "fb3d2015-bc17-4bb9-bb85-30c5cf1aaaa7"
        #    }

        postokay, response = self._fetch(
            url=url,
            payload=payload,
            content_type="application/x-www-form-urlencoded",
        )
        if not postokay:
            return False

        # Reset our token
        self.token = None

        try:
            # Extract our time from our response and subtrace 10 seconds from
            # it to give us some wiggle/grace people to re-authenticate if we
            # need to
            self.token_expiry = datetime.now() + timedelta(
                seconds=int(response.get("expires_in")) - 10
            )

        except (ValueError, AttributeError, TypeError):
            # ValueError: expires_in wasn't an integer
            # TypeError: expires_in was None
            # AttributeError: we could not extract anything from our response
            #                object.
            return False

        # Go ahead and store our token if it's available
        self.token = response.get("access_token")

        # We're authenticated
        return bool(self.token)

    def _fetch(
        self,
        url,
        payload=None,
        headers=None,
        content_type="application/json",
        method="POST",
    ):
        """Wrapper to request object."""

        # Prepare our headers:
        if not headers:
            headers = {
                "User-Agent": self.app_id,
                "Content-Type": content_type,
            }

        if self.token:
            # Are we authenticated?
            headers["Authorization"] = "Bearer " + self.token

        # Default content response object
        content = {}

        # Some Debug Logging
        self.logger.debug(
            "Office 365 %s URL:"
            f" {url} (cert_verify={self.verify_certificate})",
            method,
        )
        self.logger.debug(f"Office 365 Payload: {payload}")

        # Always call throttle before any remote server i/o is made
        self.throttle()

        # fetch function
        req = (
            requests.post
            if method == "POST"
            else (requests.put if method == "PUT" else requests.get)
        )

        try:
            r = req(
                url,
                data=(
                    json.dumps(payload)
                    if content_type and content_type.endswith("/json")
                    else payload
                ),
                headers=headers,
                verify=self.verify_certificate,
                timeout=self.request_timeout,
            )

            if r.status_code not in (
                requests.codes.ok,
                requests.codes.created,
                requests.codes.accepted,
            ):

                # We had a problem
                status_str = NotifyOffice365.http_response_code_lookup(
                    r.status_code
                )

                self.logger.warning(
                    "Failed to send Office 365 %s to {}: {}error={}.".format(
                        url, ", " if status_str else "", r.status_code
                    ),
                    method,
                )

                # A Response could look like this if a Scope element was not
                # found:
                # {
                #  "error": {
                #     "code": "MissingClaimType",
                #     "message":"The token is missing the claim type \'oid\'.",
                #     "innerError": {
                #       "oAuthEventOperationId":" 7abe20-339f-4659-9381-38f52",
                #       "oAuthEventcV": "xsOSpAHSHVm3Tp4SNH5oIA.1.1",
                #       "errorUrl": "https://url",
                #       "requestId": "2328ea-ec9e-43a8-80f4-164c",
                #       "date":"2024-12-01T02:03:13"
                #  }}
                # }

                # Error 403; the below is returned if he User.Read.All
                #           Application scope is not set and a lookup is
                #           attempted.
                # {
                #   "error": {
                #     "code": "Authorization_RequestDenied",
                #     "message":
                #        "Insufficient privileges to complete the operation.",
                #     "innerError": {
                #       "date": "2024-12-06T00:15:57",
                #       "request-id":
                #         "48fdb3e7-2f1a-4f45-a5a0-99b8b851278b",
                #         "client-request-id": "48f-2f1a-4f45-a5a0-99b8"
                #     }
                #   }
                # }

                # Another response type (error 415):
                # {
                #  "error": {
                #    "code": "RequestBodyRead",
                #    "message": "A missing or empty content type header was \
                #        found when trying to read a message. The content \
                #        type header is required.",
                #  }
                # }

                self.logger.debug(f"Response Details:\r\n{r.content}")

                # Mark our failure
                return (False, content)

            try:
                content = json.loads(r.content)

            except (AttributeError, TypeError, ValueError):
                # ValueError = r.content is Unparsable
                # TypeError = r.content is None
                # AttributeError = r is None
                content = {}

        except requests.RequestException as e:
            self.logger.warning(
                f"Exception received when sending Office 365 %s to {url}: ",
                method,
            )
            self.logger.debug(f"Socket Exception: {e!s}")

            # Mark our failure
            return (False, content)

        return (True, content)

    @property
    def url_identifier(self):
        """Returns all of the identifiers that make this URL unique from
        another simliar one.

        Targets or end points should never be identified here.
        """
        return (
            self.secure_protocol[0],
            self.source,
            self.tenant,
            self.client_id,
            self.secret,
        )

    def url(self, privacy=False, *args, **kwargs):
        """Returns the URL built dynamically based on specified arguments."""

        # Extend our parameters
        params = self.url_parameters(privacy=privacy, *args, **kwargs)

        if self.cc:
            # Handle our Carbon Copy Addresses
            params["cc"] = ",".join([
                "{}{}".format(
                    "" if not self.names.get(e) else f"{self.names[e]}:", e
                )
                for e in self.cc
            ])

        if self.bcc:
            # Handle our Blind Carbon Copy Addresses
            params["bcc"] = ",".join([
                "{}{}".format(
                    "" if not self.names.get(e) else f"{self.names[e]}:", e
                )
                for e in self.bcc
            ])

        return (
            "{schema}://{source}/{tenant}/{client_id}/{secret}"
            "/{targets}/?{params}".format(
                schema=self.secure_protocol[0],
                tenant=self.pprint(self.tenant, privacy, safe=""),
                # email does not need to be escaped because it should
                # already be a valid host and username at this point
                source=self.source,
                client_id=self.pprint(self.client_id, privacy, safe=""),
                secret=self.pprint(
                    self.secret, privacy, mode=PrivacyMode.Secret, safe=""
                ),
                targets="/".join([
                    NotifyOffice365.quote(
                        "{}{}".format("" if not e[0] else f"{e[0]}:", e[1]),
                        safe="@",
                    )
                    for e in self.targets
                ]),
                params=NotifyOffice365.urlencode(params),
            )
        )

    def __len__(self):
        """Returns the number of targets associated with this notification."""
        return len(self.targets)

    @staticmethod
    def parse_url(url):
        """Parses the URL and returns enough arguments that can allow us to re-
        instantiate this object."""

        results = NotifyBase.parse_url(url, verify_host=False)
        if not results:
            # We're done early as we couldn't load the results
            return results

        # Now make a list of all our path entries
        # We need to read each entry back one at a time in reverse order
        # where each email found we mark as a target. Once we run out
        # of targets, the presume the remainder of the entries are part
        # of the secret key (since it can contain slashes in it)
        entries = NotifyOffice365.split_path(results["fullpath"])

        # Initialize our tenant
        results["tenant"] = None

        # Initialize our email
        results["email"] = None

        # From Email
        if "from" in results["qsd"] and len(results["qsd"]["from"]):
            # Extract the sending account's information
            results["source"] = NotifyOffice365.unquote(results["qsd"]["from"])

        # If tenant is occupied, then the user defined makes up our source
        elif results["user"]:
            results["source"] = "{}@{}".format(
                NotifyOffice365.unquote(results["user"]),
                NotifyOffice365.unquote(results["host"]),
            )

        else:
            # Object ID instead of email
            results["source"] = NotifyOffice365.unquote(results["host"])

        # Tenant
        if "tenant" in results["qsd"] and len(results["qsd"]["tenant"]):
            # Extract the Tenant from the argument
            results["tenant"] = NotifyOffice365.unquote(
                results["qsd"]["tenant"]
            )

        elif entries:
            results["tenant"] = NotifyOffice365.unquote(entries.pop(0))

        # OAuth2 ID
        if "oauth_id" in results["qsd"] and len(results["qsd"]["oauth_id"]):
            # Extract the API Key from an argument
            results["client_id"] = NotifyOffice365.unquote(
                results["qsd"]["oauth_id"]
            )

        elif entries:
            # Get our client_id is the first entry on the path
            results["client_id"] = NotifyOffice365.unquote(entries.pop(0))

        #
        # Prepare our target listing
        #
        results["targets"] = []
        while entries:
            # Pop the last entry
            entry = NotifyOffice365.unquote(entries.pop(-1))

            if is_email(entry):
                # Store our email and move on
                results["targets"].append(entry)
                continue

            # If we reach here, the entry we just popped is part of the secret
            # key, so put it back
            entries.append(NotifyOffice365.quote(entry, safe=""))

            # We're done
            break

        # OAuth2 Secret
        if "oauth_secret" in results["qsd"] and len(
            results["qsd"]["oauth_secret"]
        ):
            # Extract the API Secret from an argument
            results["secret"] = NotifyOffice365.unquote(
                results["qsd"]["oauth_secret"]
            )

        else:
            # Assemble our secret key which is a combination of the host
            # followed by all entries in the full path that follow up until
            # the first email
            results["secret"] = "/".join(
                [NotifyOffice365.unquote(x) for x in entries]
            )

        # Support the 'to' variable so that we can support targets this way too
        # The 'to' makes it easier to use yaml configuration
        if "to" in results["qsd"] and len(results["qsd"]["to"]):
            results["targets"] += NotifyOffice365.parse_list(
                results["qsd"]["to"]
            )

        # Handle Carbon Copy Addresses
        if "cc" in results["qsd"] and len(results["qsd"]["cc"]):
            results["cc"] = results["qsd"]["cc"]

        # Handle Blind Carbon Copy Addresses
        if "bcc" in results["qsd"] and len(results["qsd"]["bcc"]):
            results["bcc"] = results["qsd"]["bcc"]

        return results



# BSD 2-Clause License
#
# Apprise - Push Notification Library.
# Copyright (c) 2025, Chris Caron <lead2gold@gmail.com>
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
#    this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

# API Reference: https://smsmanager.cz/api/http#send

# To use this service you will need a SMS Manager account
# You will need credits (new accounts start with a few)
#     https://smsmanager.cz
#    1. Sign up and get test credit
#    2. Generate an API key in web administration.

import requests

from ..common import NotifyType
from ..locale import gettext_lazy as _
from ..utils.parse import (
    is_phone_no,
    parse_bool,
    parse_phone_no,
    validate_regex,
)
from .base import NotifyBase


class SMSManagerGateway:
    """The different gateway values."""

    HIGH = "high"
    ECONOMY = "economy"
    LOW = "low"
    DIRECT = "direct"


# Used for verification purposes
SMS_MANAGER_GATEWAYS = (
    SMSManagerGateway.HIGH,
    SMSManagerGateway.ECONOMY,
    SMSManagerGateway.LOW,
    SMSManagerGateway.DIRECT,
)


class NotifySMSManager(NotifyBase):
    """A wrapper for SMS Manager Notifications."""

    # The default descriptive name associated with the Notification
    service_name = "SMS Manager"

    # The services URL
    service_url = "https://smsmanager.cz"

    # All notification requests are secure
    secure_protocol = (
        "smsmgr",
        "smsmanager",
    )

    # A URL that takes you to the setup/help of the specific protocol
    setup_url = "https://github.com/caronc/apprise/wiki/Notify_sms_manager"

    # SMS Manager uses the http protocol with JSON requests
    notify_url = "https://http-api.smsmanager.cz/Send"

    # The maximum amount of texts that can go out in one batch
    default_batch_size = 4000

    # The maximum length of the body
    body_maxlen = 160

    # A title can not be used for SMS Messages.  Setting this to zero will
    # cause any title (if defined) to get placed into the message body.
    title_maxlen = 0

    # Define object templates
    templates = ("{schema}://{apikey}@{targets}",)

    # Define our template tokens
    template_tokens = dict(
        NotifyBase.template_tokens,
        **{
            "apikey": {
                "name": _("API Key"),
                "type": "string",
                "private": True,
                "required": True,
            },
            "target_phone": {
                "name": _("Target Phone No"),
                "type": "string",
                "prefix": "+",
                "regex": (r"^[0-9\s)(+-]+$", "i"),
                "map_to": "targets",
            },
            "targets": {
                "name": _("Targets"),
                "type": "list:string",
                "required": True,
            },
        },
    )

    # Define our template arguments
    template_args = dict(
        NotifyBase.template_args,
        **{
            "key": {
                "alias_of": "apikey",
            },
            "to": {
                "alias_of": "targets",
            },
            "from": {
                "name": _("From Phone No"),
                "type": "string",
                "regex": (r"^\+?[0-9\s)(+-]+$", "i"),
                "map_to": "sender",
            },
            "sender": {
                "alias_of": "from",
            },
            "gateway": {
                "name": _("Gateway"),
                "type": "choice:string",
                "values": SMS_MANAGER_GATEWAYS,
                "default": SMS_MANAGER_GATEWAYS[0],
            },
            "batch": {
                "name": _("Batch Mode"),
                "type": "bool",
                "default": False,
            },
        },
    )

    def __init__(
        self,
        apikey=None,
        sender=None,
        targets=None,
        batch=None,
        gateway=None,
        **kwargs,
    ):
        """Initialize SMS Manager Object."""
        super().__init__(**kwargs)

        self.apikey = validate_regex(apikey)
        if not self.apikey:
            msg = f"An invalid API Key ({apikey}) was specified."
            self.logger.warning(msg)
            raise TypeError(msg)

        # Setup our gateway
        self.gateway = (
            self.template_args["gateway"]["default"]
            if not isinstance(gateway, str)
            else gateway.lower()
        )
        if self.gateway not in SMS_MANAGER_GATEWAYS:
            msg = f"The Gateway specified ({gateway}) is invalid."
            self.logger.warning(msg)
            raise TypeError(msg)

        # Define whether or not we should operate in a batch mode
        self.batch = (
            self.template_args["batch"]["default"]
            if batch is None
            else bool(batch)
        )

        # Maximum 11 characters and must be approved by administrators of site
        self.sender = sender[0:11] if isinstance(sender, str) else None

        # Parse our targets
        self.targets = []

        for target in parse_phone_no(targets):
            # Parse each phone number we found
            # It is documented that numbers with a length of 9 characters are
            # supplemented by "420".
            result = is_phone_no(target, min_len=9)
            if result:
                # Carry forward '+' if defined, otherwise do not...
                self.targets.append(
                    ("+" + result["full"])
                    if target.lstrip()[0] == "+"
                    else result["full"]
                )
                continue

            self.logger.warning(
                f"Dropped invalid phone # ({target}) specified.",
            )

        return

    def send(self, body, title="", notify_type=NotifyType.INFO, **kwargs):
        """Perform SMS Manager Notification."""

        if not self.targets:
            # We have nothing to notify
            self.logger.warning("There are no SMS Manager targets to notify")
            return False

        # error tracking (used for function return)
        has_error = False

        # Send in batches if identified to do so
        batch_size = 1 if not self.batch else self.default_batch_size

        # Prepare our headers
        headers = {
            "User-Agent": self.app_id,
        }

        # Prepare our targets
        targets = (
            list(self.targets)
            if batch_size == 1
            else [
                self.targets[index : index + batch_size]
                for index in range(0, len(self.targets), batch_size)
            ]
        )

        while len(targets):
            # Get our target to notify
            target = targets.pop(0)

            # Prepare our payload
            # Note: Payload is assembled inside of our while-loop due to
            #       mock testing issues (payload singleton isn't persistent
            #       when performing follow up checks on the params object.
            payload = {
                "apikey": self.apikey,
                "gateway": self.gateway,
                # The number gets populated in the loop below
                "number": None,
                "message": body,
            }

            if self.sender:
                # Sender is ony set if specified
                payload["sender"] = self.sender

            # Printable target details
            if isinstance(target, list):
                p_target = f"{len(target)} targets"

                # Prepare our target numbers
                payload["number"] = ";".join(target)

            else:
                p_target = target
                # Prepare our target numbers
                payload["number"] = target

            # Some Debug Logging
            self.logger.debug(
                "SMS Manager POST URL:"
                f" {self.notify_url} (cert_verify={self.verify_certificate})"
            )
            self.logger.debug(f"SMS Manager Payload: {payload}")

            # Always call throttle before any remote server i/o is made
            self.throttle()
            try:
                r = requests.get(
                    self.notify_url,
                    params=payload,
                    headers=headers,
                    verify=self.verify_certificate,
                    timeout=self.request_timeout,
                )

                if r.status_code != requests.codes.ok:
                    # We had a problem
                    status_str = NotifyBase.http_response_code_lookup(
                        r.status_code
                    )

                    # set up our status code to use
                    status_code = r.status_code

                    self.logger.warning(
                        "Failed to send SMS Manager notification to {}: "
                        "{}{}error={}.".format(
                            p_target,
                            status_str,
                            ", " if status_str else "",
                            status_code,
                        )
                    )

                    self.logger.debug(f"Response Details:\r\n{r.content}")

                    # Mark our failure
                    has_error = True
                    continue

                else:
                    self.logger.info(
                        f"Sent SMS Manager notification to {p_target}."
                    )

            except requests.RequestException as e:
                self.logger.warning(
                    "A Connection error occurred sending SMS Manager: to %s ",
                    p_target,
                )
                self.logger.debug(f"Socket Exception: {e!s}")

                # Mark our failure
                has_error = True
                continue

        return not has_error

    @property
    def url_identifier(self):
        """Returns all of the identifiers that make this URL unique from
        another simliar one.

        Targets or end points should never be identified here.
        """
        return (self.secure_protocol[0], self.apikey)

    def url(self, privacy=False, *args, **kwargs):
        """Returns the URL built dynamically based on specified arguments."""

        # Define any URL parameters
        params = {
            "batch": "yes" if self.batch else "no",
            "gateway": self.gateway,
        }

        if self.sender:
            # Set our sender if it was set
            params["sender"] = self.sender

        # Extend our parameters
        params.update(self.url_parameters(privacy=privacy, *args, **kwargs))

        return "{schema}://{apikey}@{targets}?{params}".format(
            schema=self.secure_protocol[0],
            apikey=self.pprint(self.apikey, privacy, safe=""),
            targets="/".join([
                NotifySMSManager.quote(f"{x}", safe="+") for x in self.targets
            ]),
            params=NotifySMSManager.urlencode(params),
        )

    def __len__(self):
        """Returns the number of targets associated with this notification."""

        #
        # Factor batch into calculation
        #
        # Note: Groups always require a separate request (and can not be
        # included in batch calculations)
        batch_size = 1 if not self.batch else self.default_batch_size
        targets = len(self.targets)
        if batch_size > 1:
            targets = int(targets / batch_size) + (
                1 if targets % batch_size else 0
            )

        return targets

    @staticmethod
    def parse_url(url):
        """Parses the URL and returns enough arguments that can allow us to re-
        instantiate this object."""

        results = NotifyBase.parse_url(url, verify_host=False)
        if not results:
            # We're done early as we couldn't load the results
            return results

        # Get our API Key
        results["apikey"] = NotifySMSManager.unquote(results["user"])

        # Store our targets
        results["targets"] = [
            *NotifySMSManager.parse_phone_no(results["host"]),
            *NotifySMSManager.split_path(results["fullpath"]),
        ]

        # The 'from' makes it easier to use yaml configuration
        if "from" in results["qsd"] and len(results["qsd"]["from"]):
            results["sender"] = NotifySMSManager.unquote(
                results["qsd"]["from"]
            )

        elif "sender" in results["qsd"] and len(results["qsd"]["sender"]):
            # Support sender= value as well to align with SMS Manager API
            results["sender"] = NotifySMSManager.unquote(
                results["qsd"]["sender"]
            )

        # Support the 'to' variable so that we can support targets this way too
        # The 'to' makes it easier to use yaml configuration
        if "to" in results["qsd"] and len(results["qsd"]["to"]):
            results["targets"] += NotifySMSManager.parse_phone_no(
                results["qsd"]["to"]
            )

        if "key" in results["qsd"] and len(results["qsd"]["key"]):
            results["apikey"] = NotifySMSManager.unquote(results["qsd"]["key"])

        # Get Batch Mode Flag
        results["batch"] = parse_bool(
            results["qsd"].get(
                "batch", NotifySMSManager.template_args["batch"]["default"]
            )
        )

        # Define our gateway
        if "gateway" in results["qsd"] and len(results["qsd"]["gateway"]):
            results["gateway"] = NotifySMSManager.unquote(
                results["qsd"]["gateway"]
            )

        return results
